Set Implicit Arguments.
From DEV Require Export LibCore.
From DEV Require Export LibSepTLCbuffer LibSepVar LibSepFmap.
From DEV Require LibLogic.
Module Fmap := LibSepFmap.

(* ########################################################### *)
(** ** Syntax *)

(** The grammar of the deeply embedded language includes terms and
    values. Values include basic values such as [int] and [bool],
    locations (memory addresses, represented by natural numbers),
    and primitive operations. *)

(** The grammar of primitive operations includes just a few operations. *)

Inductive prim : Type :=
  | val_ref : prim
  | val_get : prim
  | val_set : prim
  | val_free : prim
  | val_rand : prim
  | val_add : prim.

Definition loc : Type := nat.

Definition null : loc := 0%nat.

(** The grammar of values. *)

Inductive val : Type :=
  | val_unit : val
  | val_bool : bool -> val
  | val_int : int -> val
  | val_loc : loc -> val
  | val_prim : prim -> val
  | val_fix : var -> var -> trm -> val
  | val_error : val

(** The grammar of terms. *)

with trm : Type :=
  | trm_val : val -> trm
  | trm_var : var -> trm
  | trm_fix : var -> var -> trm -> trm
  | trm_app : trm -> trm -> trm
  | trm_let : var -> trm -> trm -> trm
  | trm_if : trm -> trm -> trm -> trm.

(** A state consists of a finite map from location to values.
    Records and arrays are represented as sets of consecutive cells,
    preceeded by a header field describing the length of the block. *)

Definition state : Type := fmap loc val.

(** [state_empty] is a handy notation to avoid providing
    type arguments to [Fmap.empty] *)

Notation "'state_empty'" := (@Fmap.empty loc val)
  (at level 0).

(** [h1 \u h2] is an optional notation for union of two states;
    in this file, it is only used for pretty-printing. *)

Notation "h1 \u h2" := (Fmap.union h1 h2)
  (at level 37, right associativity).

(** Implicit types associated with meta-variables. *)

Implicit Types f : var.
Implicit Types b : bool.
Implicit Types p : loc.
Implicit Types n : int.
Implicit Types v w r vf vx : val.
Implicit Types t : trm.
Implicit Types s : state.

(** The types of values and heap values are inhabited. *)

Global Instance Inhab_val : Inhab val.
Proof using. apply (Inhab_of_val val_unit). Qed.

(** Coercions to improve conciseness in the statment of evaluation rules. *)

Coercion val_bool : bool >-> val.
Coercion val_int : Z >-> val.
Coercion val_loc : loc >-> val.
Coercion val_prim : prim >-> val.

Coercion trm_val : val >-> trm.
Coercion trm_var : var >-> trm.
Coercion trm_app : trm >-> Funclass.


(* ########################################################### *)
(** ** Substitution *)

(** Capture-avoiding substitution, standard definition. *)

Fixpoint subst (y:var) (v':val) (t:trm) : trm :=
  let aux t := subst y v' t in
  let if_y_eq x t1 t2 := if var_eq x y then t1 else t2 in
  match t with
  | trm_val v => trm_val v
  | trm_var x => if_y_eq x (trm_val v') t
  | trm_fix f x t1 => trm_fix f x (if_y_eq f t1 (if_y_eq x t1 (aux t1)))
  | trm_app t1 t2 => trm_app (aux t1) (aux t2)
  | trm_let x t1 t2 => trm_let x (aux t1) (if_y_eq x t2 (aux t2))
  | trm_if t0 t1 t2 => trm_if (aux t0) (aux t1) (aux t2)
  end.


(* ########################################################### *)
(** ** Postconditions *)

Declare Scope heap_scope.
Open Scope heap_scope.

(** Heap predicate *)

Definition hprop := state -> Prop.

Implicit Types H : hprop.
Implicit Types Q : val->hprop.

(** Heap entailment *)

Definition himpl (H1 H2:hprop) : Prop :=
  forall h, H1 h -> H2 h.

Notation "H1 ==> H2" := (himpl H1 H2)
  (at level 55) : heap_scope.

(** Postcondition entailment *)

Definition qimpl A (Q1 Q2:A->hprop) :=
  forall x, himpl (Q1 x) (Q2 x).

Notation "Q1 ===> Q2" := (qimpl Q1 Q2)
  (at level 55) : heap_scope.

(** Properties *)

Lemma himpl_refl : forall H,
  H ==> H.
Proof using. introv M. auto. Qed.

Hint Resolve himpl_refl.

Lemma himpl_trans : forall H2 H1 H3,
  (H1 ==> H2) ->
  (H2 ==> H3) ->
  (H1 ==> H3).
Proof using. introv M1 M2. unfolds* himpl. Qed.

Lemma qimpl_refl : forall A (Q:A->hprop),
  Q ===> Q.
Proof using. intros. unfolds*. Qed.

Hint Resolve qimpl_refl.


(* ########################################################### *)
(** ** The Always-True and Always-False Postconditions *)

(** Let [Any] denotes the postcondition that accepts any result. *)

Definition Any : val->state->Prop :=
  fun v s => True.

Hint Unfold Any.

(** Let [Empty] denotes the postcondition that rejects any result. *)

Definition Empty : val->state->Prop :=
  fun v s => False.

Hint Unfold Empty.


(* ########################################################### *)
(* ########################################################### *)
(* ########################################################### *)
(** * Standard semantics *)

(* ########################################################### *)
(** ** Standard Big-step *)

(** Judgment [big s t s' v]. *)

Inductive big : state -> trm -> state -> val -> Prop :=
  | big_val : forall s v,
      big s (trm_val v) s v
  | big_fix : forall s f x t1,
      big s (trm_fix f x t1) s (val_fix f x t1)
  | big_app_fix : forall s1 s2 v1 v2 f x t1 v,
      v1 = val_fix f x t1 ->
      big s1 (subst x v2 (subst f v1 t1)) s2 v ->
      big s1 (trm_app v1 v2) s2 v
  | big_let : forall s1 s2 s3 x t1 t2 v1 r,
      big s1 t1 s2 v1 ->
      big s2 (subst x v1 t2) s3 r ->
      big s1 (trm_let x t1 t2) s3 r
  | big_if : forall s1 s2 b v t1 t2,
      big s1 (if b then t1 else t2) s2 v ->
      big s1 (trm_if (val_bool b) t1 t2) s2 v
  | big_add : forall s n1 n2,
      big s (val_add (val_int n1) (val_int n2)) s (val_int (n1 + n2))
  | big_rand : forall s n n1,
      0 <= n1 < n ->
      big s (val_rand (val_int n)) s (val_int n1)
  | big_ref : forall s v p,
      ~ Fmap.indom s p ->
      big s (val_ref v) (Fmap.update s p v) (val_loc p)
  | big_get : forall s p,
      Fmap.indom s p ->
      big s (val_get (val_loc p)) s (Fmap.read s p)
  | big_set : forall s p v,
      Fmap.indom s p ->
      big s (val_set (val_loc p) v) (Fmap.update s p v) val_unit
  | big_free : forall s p,
      Fmap.indom s p ->
      big s (val_free (val_loc p)) (Fmap.remove s p) val_unit.


(* ########################################################### *)
(** ** Coinductive Big-step *)

(** [codiv s t] asserts that there exists a diverging execution
    of the program [(s,t)]. This judgment is defined coinductively,
    and depends on the judgment [big]. *)

CoInductive codiv : state -> trm -> Prop :=
  | codiv_app_fix : forall s1 v1 v2 f x t1,
      v1 = val_fix f x t1 ->
      codiv s1 (subst x v2 (subst f v1 t1)) ->
      codiv s1 (trm_app v1 v2)
  | codiv_let_ctx : forall s1 x t1 t2,
      codiv s1 t1 ->
      codiv s1 (trm_let x t1 t2)
  | codiv_let : forall s1 s2 x t1 t2 v1,
      big s1 t1 s2 v1 ->
      codiv s2 (subst x v1 t2) ->
      codiv s1 (trm_let x t1 t2)
  | codiv_if : forall s1 b t1 t2,
      codiv s1 (if b then t1 else t2) ->
      codiv s1 (trm_if (val_bool b) t1 t2).


(* ########################################################### *)
(** ** Standard Small-Step *)

(** The judgment [step s t s' t'] describes the small-step reduction relation:
    it asserts that the program configuration [(s,t)] can take one reduction
    step towards the program configuration [(s',t')]. *)

Inductive step : state -> trm -> state -> trm -> Prop :=

  (* Unique context rule *)
  | step_let_ctx : forall s1 s2 x t1 t1' t2,
      step s1 t1 s2 t1' ->
      step s1 (trm_let x t1 t2) s2 (trm_let x t1' t2)

  (* Beta reductions *)
  | step_fix : forall s f x t1,
      step s (trm_fix f x t1) s (val_fix f x t1)
  | step_app_fix : forall s v1 v2 f x t1,
      v1 = val_fix f x t1 ->
      step s (trm_app v1 v2) s (subst x v2 (subst f v1 t1))
  | step_if : forall s b t1 t2,
      step s (trm_if (val_bool b) t1 t2) s (if b then t1 else t2)
  | step_let : forall s x t2 v1,
      step s (trm_let x v1 t2) s (subst x v1 t2)

  (* Primitive operations *)
  | step_add : forall s n1 n2,
      step s (val_add (val_int n1) (val_int n2)) s (val_int (n1 + n2))
  | step_rand : forall s n n1,
      0 <= n1 < n ->
      step s (val_rand (val_int n)) s (val_int n1)
  | step_ref : forall s v p,
      ~ Fmap.indom s p ->
      step s (val_ref v) (Fmap.update s p v) (val_loc p)
  | step_get : forall s p,
      Fmap.indom s p ->
      step s (val_get (val_loc p)) s (Fmap.read s p)
  | step_set : forall s p v,
      Fmap.indom s p ->
      step s (val_set (val_loc p) v) (Fmap.update s p v) val_unit
  | step_free : forall s p,
      Fmap.indom s p ->
      step s (val_free (val_loc p)) (Fmap.remove s p) val_unit.

(** The judgment [steps s t s' t'] corresponds to the reflexive-transitive
    closure of [step]. Concretely, this judgment asserts that the configuration
    [(s,t)] can reduce in zero, one, or several steps to [(s',t')]. *)

Inductive steps : state -> trm -> state -> trm -> Prop :=
  | steps_refl : forall s t,
      steps s t s t
  | steps_step : forall s1 s2 s3 t1 t2 t3,
      step s1 t1 s2 t2 ->
      steps s2 t2 s3 t3 ->
      steps s1 t1 s3 t3.


(* ########################################################### *)
(** ** Properties of the Small Step Judgment *)

(** Values can't step. *)

Lemma step_val_inv : forall s v s' t',
  step s (trm_val v) s' t' ->
  False.
Proof using. introv S. inverts S; false_invert. Qed.

(** Reformulation of the above using existentials. *)

Lemma exists_step_val_inv : forall s v,
  (exists s' t', step s v s' t') ->
  False.
Proof using. introv (s'&t'&S). applys* step_val_inv S. Qed.


(* ########################################################### *)
(** ** Equivalence for Between Standard Small-Step and Standard Big-Step *)

(** In this section, we establish the equivalence between
    the big-step judgment [big s t s' v] and the iterated
    small-step judgment [steps s t s' (trm_val v)]. *)

(** Consider first the direction from small-step to big-step.
    The proof involves a key lemma, which shows how to "glue"
    a one-step reduction with a big-step evaluation to produce
    a (larger) big-step evaluation. *)

Lemma big_of_step_and_eval : forall s1 s2 t1 t2 s3 v,
  step s1 t1 s2 t2 ->
  big s2 t2 s3 v ->
  big s1 t1 s3 v.
Proof using.
  introv M1 M2. gen s3 v. induction M1; intros.
  { inverts M2 as; try false_invert.
    { introv M3 M4. applys* big_let. } }
  { inverts M2; try false_invert. applys big_fix. }
  { applys* big_app_fix. }
  { applys* big_if. }
  { applys* big_let. applys big_val. }
  { inverts M2; try false_invert. applys* big_add. }
  { inverts M2; try false_invert. applys* big_rand. }
  { inverts M2; try false_invert. applys* big_ref. }
  { inverts M2; try false_invert. applys* big_get. }
  { inverts M2; try false_invert. applys* big_set. }
  { inverts M2; try false_invert. applys* big_free. }
Qed.

(** Using the above lemmas, to establish the implication between
    the iterated small-step relation, it remains to perform an
    induction over the sequence of individual steps involved. *)

Lemma big_of_steps : forall s1 s2 t v,
  steps s1 t s2 (trm_val v) ->
  big s1 t s2 v.
Proof.
  introv M. gen_eq t': (trm_val v). gen v.
  induction M; intros; subst.
  { applys* big_val. }
  { applys* big_of_step_and_eval. }
Qed.

(** Let's now tackle the reciprocal direction, from big-step to
    small-step. This time, the key auxiliary lemma is a context
    rule for describing how a sequence of steps can be applied
    under a let-context. *)

Lemma steps_let : forall s1 s2 s3 v1 v3 x t1 t2,
  steps s1 t1 s2 (trm_val v1) ->
  steps s2 (subst x v1 t2) s3 v3 ->
  steps s1 (trm_let x t1 t2) s3 v3.
Proof using.
  introv M1 M2. gen_eq t1': (trm_val v1). gen v1.
  induction M1; intros; subst.
  { applys steps_step. { applys step_let. } { applys M2. } }
  { rename H into R1. applys steps_step.
    { applys step_let_ctx R1. }
    { applys* IHM1 M2. } }
Qed.

(** There remains to perfom an induction over the big-step
    evaluation relation. The only nontrivial case is that of
    the let-binding, for which we need to exploit [steps_let]. *)

Lemma steps_of_eval : forall s1 s2 t v,
  big s1 t s2 v ->
  steps s1 t s2 (trm_val v).
Proof using.
  introv M. induction M.
  { applys steps_refl. }
  { applys steps_step. { applys step_fix. } { applys steps_refl. } }
  { applys steps_step. { applys* step_app_fix. } { applys IHM. } }
  { applys* steps_let IHM1 IHM2. }
  { applys steps_step. { applys* step_if. } { applys IHM. } }
  { applys steps_step. { applys* step_add. } { applys steps_refl. } }
  { applys steps_step. { applys* step_rand. } { applys steps_refl. } }
  { applys steps_step. { applys* step_ref. } { applys steps_refl. } }
  { applys steps_step. { applys* step_get. } { applys steps_refl. } }
  { applys steps_step. { applys* step_set. } { applys steps_refl. } }
  { applys steps_step. { applys* step_free. } { applys steps_refl. } }
Qed.

(** Putting the two directions together yields the desired equivalence. *)

Lemma big_iff_steps : forall s t s' v,
  big s t s' v <-> steps s t s' (trm_val v).
Proof using.
  iff M.
  { applys* steps_of_eval. }
  { applys* big_of_steps. }
Qed.


(* ########################################################### *)
(* ########################################################### *)
(* ########################################################### *)
(** * Common definitions  *)

(* ########################################################### *)
(** ** Definition of [strongestpost] and [exactly] *)

(** Let [strongestpost s t] denote the strongest postcondition
    (strongest-post), that is, [fun v s' => big s t s' v]. *)

Definition strongestpost (s:state) (t:trm) : val->hprop :=
  fun v s' => big s t s' v.

(* The postcondition [exactly v s'] to assert that the output of a program
   is exactly the value [v] in the state [s']. This postcondition
   corresponds to a singleton set. *)

Definition exactly (v:val) (s':state) : val->hprop :=
  fun v0 s0 => v0 = v /\ s0 = s'.

(** By definition, the postcondition [exactly v s'] holds only of [v]
    and [s'], thus the only introduction rule for [exactly] is the
    following reflexivity rule. *)

Lemma exactly_intro : forall v s',
  exactly v s' v s'.
Proof using. intros. hnfs*. Qed.



(* ########################################################### *)
(* ########################################################### *)
(* ########################################################### *)
(** * Inductive Omni-Big-Step *)

(* ########################################################### *)
(** ** Definition of Omni-Big-Step *)

(** Judgment [omnibig s t Q] *)

Inductive omnibig : state -> trm -> (val->state->Prop) -> Prop :=
  | omnibig_val : forall s v Q,
      Q v s ->
      omnibig s (trm_val v) Q
  | omnibig_fix : forall s f x t1 Q,
      Q (val_fix f x t1) s ->
      omnibig s (trm_fix f x t1) Q
  | omnibig_app_fix : forall s v1 v2 f x t1 Q,
      v1 = val_fix f x t1 ->
      omnibig s (subst x v2 (subst f v1 t1)) Q ->
      omnibig s (trm_app v1 v2) Q
  | omnibig_let : forall Q1 s x t1 t2 Q,
      omnibig s t1 Q1 ->
      (forall v1 s2, Q1 v1 s2 -> omnibig s2 (subst x v1 t2) Q) ->
      omnibig s (trm_let x t1 t2) Q
  | omnibig_if : forall s b t1 t2 Q,
      omnibig s (if b then t1 else t2) Q ->
      omnibig s (trm_if (val_bool b) t1 t2) Q
  | omnibig_add : forall s n1 n2 Q,
      Q (val_int (n1 + n2)) s ->
      omnibig s (val_add (val_int n1) (val_int n2)) Q
  | omnibig_rand : forall s n Q,
      n > 0 ->
      (forall n1, 0 <= n1 < n -> Q n1 s) ->
      omnibig s (val_rand (val_int n)) Q
  | omnibig_ref : forall s v Q,
      (forall p, ~ Fmap.indom s p ->
         Q (val_loc p) (Fmap.update s p v)) ->
      omnibig s (val_ref v) Q
  | omnibig_get : forall s p Q,
      Fmap.indom s p ->
      Q (Fmap.read s p) s ->
      omnibig s (val_get (val_loc p)) Q
  | omnibig_set : forall s p v Q,
      Fmap.indom s p ->
      Q val_unit (Fmap.update s p v) ->
      omnibig s (val_set (val_loc p) v) Q
  | omnibig_free : forall s p Q,
      Fmap.indom s p ->
      Q val_unit (Fmap.remove s p) ->
      omnibig s (val_free (val_loc p)) Q.

(** [omnibig] is covariant in the postcondition (omni-big-conseqence) *)

Lemma omnibig_conseq : forall s t Q1 Q2,
  omnibig s t Q1 ->
  Q1 ===> Q2 ->
  omnibig s t Q2.
Proof using.
  introv M W.
  asserts W': (forall v h, Q1 v h -> Q2 v h). { auto. } clear W.
  induction M; try solve [ constructors* ].
Qed.


(* ########################################################### *)
(** ** Relationship with WP Formulation *)

(** The rule [omnibig_let] shares similarities with the statement of
    the weakest-precondition style reasoning rule for let-bindings.
    Prove the following alternative statement, with a wp-style flavor. *)

Lemma omnibig_let' : forall s1 x t1 t2 Q,
  omnibig s1 t1 (fun v1 s2 => omnibig s2 (subst x v1 t2) Q) ->
  omnibig s1 (trm_let x t1 t2) Q.
Proof using.
  introv M. applys* omnibig_let M.
Qed.

(** Reciprocally, prove that the statement considered in the inductive
    definition of [omnibig] is derivable from [omnibig_let']. More precisely,
    prove the statement below by using [omnibig_let'] and [omnibig_conseq]. *)

Lemma omnibig_let_of_omnibig_let' : forall Q1 s1 x t1 t2 Q,
  omnibig s1 t1 Q1 ->
  (forall v1 s2, Q1 v1 s2 -> omnibig s2 (subst x v1 t2) Q) ->
  omnibig s1 (trm_let x t1 t2) Q.
Proof using.
  introv M1 M2. applys omnibig_let'. applys omnibig_conseq M1.
  intros v1 s2 K. applys M2 K.
Qed.

(** One way wonder whether we could have used the wp-style formulation
    of the semantics of let-bindings directly in the definition of [omnibig].
    The answer is negative. Doing so would lead to an invalid inductive
    definition, involving a "non strictly positive occurrence". To check it
    out, uncomment the definitions below to observe Coq's complaint.
[[
  Inductive omnibig' : state -> trm -> (val->state->Prop) -> Prop :=
    | omnibig'_let : forall s1 x t1 t2 Q,
       omnibig' s1 t1 (fun v1 s2 => omnibig' s2 (subst x v1 t2) Q) ->
       omnibig' s1 (trm_let x t1 t2) Q.

  Inductive omnibig' : state -> trm -> (val->state->Prop) -> Prop :=
    | omnibig'_let : forall Q1 s1 x t1 t2 Q,
       omnibig' s1 t1 Q1 ->
       Q1 = (fun v1 s2 => omnibig' s2 (subst x v1 t2) Q) ->
       omnibig' s1 (trm_let x t1 t2) Q.
]]

   Yet another unsuccessful try:
[[
  Inductive omnibig' : state -> trm -> (val->state->Prop) -> Prop :=
    | omnibig'_let : forall Q1 s1 x t1 t2 Q,
       omnibig' s1 t1 Q1 ->
       (forall v1 s2, Q1 v1 s2 -> exists Q2, omnibig' s2 (subst x v1 t2) Q2) ->
       Q = (fun v2 s3 => exists v1 s2, Q1 v1 s2 /\ exists Q2, Q2 v2 s3 /\ omnibig' s2 (subst x v1 t2) Q2) ->
       omnibig' s1 (trm_let x t1 t2) Q.
]]
*)


(* ########################################################### *)
(** ** Properties of Omni-Big-Step, Part 1 *)

(** Assume [omnibig s t Q] to hold. Prove that the postcondition [Q] holds of
    any output that [(s,t)] may evaluate to according to the judgment [big]. *)

Lemma omnibig_and_big_inv : forall s t v s' Q,
  omnibig s t Q ->
  big s t s' v ->
  Q v s'.
Proof using.
  introv M R. gen v s'.
  induction M; intros; try solve [inverts* R; tryfalse].
  { invert R; try solve [intros; false].
    introv E R. intros. subst. inverts E. applys* IHM. }
Qed.

(** If [omnibig s t Q] holds, then there exists an output
    [(s',v)] that [(s,t)] may evaluate to according to the judgment
    [big], and that this output satisfies [Q]. *)

Lemma omnibig_to_one_big : forall s t Q,
  omnibig s t Q ->
  exists s' v, big s t s' v /\ Q v s'.
Proof using.
  introv M. cuts (s'&v&R): (exists s' v, big s t s' v).
  { exists. split. { applys R. } { applys omnibig_and_big_inv M R. } }
  induction M.
  { exists. applys big_val. }
  { exists. applys big_fix. }
  { rename IHM into IHM1, M into M1.
    forwards (s'&v&R): IHM1.
    exists. applys* big_app_fix R. }
  { rename IHM into IHM1, H0 into IHM2, M into M1, H into M2.
    forwards (s1'&v1&R1): IHM1.
    lets HQ1: omnibig_and_big_inv M1 R1.
    forwards (s'&v&R2): IHM2 HQ1.
    exists. applys big_let R1 R2. }
  { forwards (s'&v&R1): IHM.
    exists. applys* big_if R1. }
  { exists. applys* big_add. }
  { rename H0 into N.
    exists. applys* big_rand 0. math. }
  { forwards* (p&D&N): (exists_fresh null s).
    exists. applys* big_ref. }
  { exists. applys* big_get. }
  { exists. applys* big_set. }
  { exists. applys* big_free. }
Qed.


(* ########################################################### *)
(** ** Properties of Omni-Big-Step, Part 2 *)

(** A special case is that of programs that admit a unique possible
    execution, that is, of programs that are deterministic even though
    the host language includes potentially non-deterministic features.
    Such a deterministic program admits a unique output. *)

(** A program [(s,t)] admits a unique execution with [(s',v)] as output
    if and only if [omnibig s t (exactly s' v)] holds. This same property
    can be expressed in terms of the judgment [big], by asserting that
    [big s t v2 s2] holds if and only if [v2 = v] and [s2 = s']. Let us
    prove the two characterizations equivalent. *)

(** We prove that if [omnibig s t Q] holds for a singleton set [Q],
    then the big-step judgment holds for that unique result. *)

Lemma big_of_omnibig_exactly : forall s t v s',
  omnibig s t (exactly v s') ->
  big s t s' v.
Proof using.
  introv M. lets (s2&v2&R&E): omnibig_to_one_big M.
  unfolds exactly. destruct E as (->&->). applys R.
Qed.

(** A slightly stronger result, not needed in the other proofs. *)

Lemma evalntb_exactly_iff_big_exactly : forall s t v s',
  omnibig s t (exactly v s') ->
  (forall s2 v2, big s t s2 v2 <-> (v2 = v /\ s2 = s')).
Proof using.
  introv M. iff R (->&->).
  { applys omnibig_and_big_inv M R. }
  { lets (s2&v2&R&E): omnibig_to_one_big M.
    unfolds exactly. destruct E as (->&->). applys R. }
Qed.


(* ########################################################### *)
(** ** Properties of Omni-Big-Step, Part 3 *)

(** If [omnibig s t Q] holds for some [Q], then
    [omnibig s t (strongestpost s t)] holds.
    See also lemma [omnibig_strongestpost_of_terminates]. *)

Lemma omnibig_strongestpost_of_omnibig : forall s t Q,
  omnibig s t Q ->
  omnibig s t (strongestpost s t).
Proof using.
  introv M. unfold strongestpost. induction M.
  { applys omnibig_val. applys big_val. }
  { applys omnibig_fix. applys big_fix. }
  { applys* omnibig_app_fix. applys omnibig_conseq IHM.
    { intros v s' R. applys* big_app_fix R. } }
  { rename IHM into IHM1, H0 into IHM2.
    applys omnibig_let (fun v1 s' => big s t1 s' v1 /\ Q1 v1 s').
    { applys omnibig_conseq IHM1. { intros v1 s' R. split.
      { applys R. }
      { applys omnibig_and_big_inv M R. } } }
    { intros v1 s2 (R1&K). applys omnibig_conseq. applys IHM2 K.
      { intros v s' R2. applys big_let R1 R2. } } }
  { applys omnibig_if. applys omnibig_conseq IHM.
    { intros v s' R. applys big_if R. } }
  { applys omnibig_add. applys big_add. }
  { applys* omnibig_rand. intros n1 N1. applys* big_rand. }
  { applys* omnibig_ref. intros p Hp. applys* big_ref. }
  { applys* omnibig_get. applys* big_get. }
  { applys* omnibig_set. applys* big_set. }
  { applys* omnibig_free. applys* big_free. }
Qed.

(** Reciprocally, the strongest postcondition is the smallest
    possible postcondition for the judgment (strongest-post-minimal). *)

Lemma omnibig_inv_strongestpost_qimpl : forall s t Q,
  omnibig s t Q ->
  strongestpost s t ===> Q.
Proof using.
  introv M. unfold strongestpost. intros v s' K. applys omnibig_and_big_inv M K.
Qed.


(* ########################################################### *)
(** ** Interpretation of Omni-Big-Step w.r.t. [terminates] *)

(** The predicate [terminates s t] asserts that all executions of the
    configuration [t/s] terminate---none of them diverges or get
    stuck. Its definition is a simplified version of [omnibig] where all
    occurences of [Q] are removed. In the rule for let-bindings,
    namely [terminates_let], the quantification over the configuration
    [v1/s2] is done by refering to the big-step judgment [big]. *)

Inductive terminates : state -> trm -> Prop :=
  | terminates_val : forall s v,
      terminates s (trm_val v)
  | terminates_fix : forall s f x t1,
      terminates s (trm_fix f x t1)
  | terminates_app_fix : forall s v1 v2 f x t1,
      v1 = val_fix f x t1 ->
      terminates s (subst x v2 (subst f v1 t1)) ->
      terminates s (trm_app v1 v2)
  | terminates_let : forall s x t1 t2,
      terminates s t1 ->
      (forall v1 s2, big s t1 s2 v1 -> terminates s2 (subst x v1 t2)) ->
      terminates s (trm_let x t1 t2)
  | terminates_if : forall s b t1 t2,
      terminates s (if b then t1 else t2) ->
      terminates s (trm_if (val_bool b) t1 t2)
  | terminatess_add : forall s n1 n2,
      terminates s (val_add (val_int n1) (val_int n2))
  | terminates_rand : forall s n,
      n > 0 ->
      terminates s (val_rand (val_int n))
  | terminates_ref : forall s v,
      terminates s (val_ref v)
  | terminates_get : forall s p,
      Fmap.indom s p ->
      terminates s (val_get (val_loc p))
  | terminates_set : forall s p v,
      Fmap.indom s p ->
      terminates s (val_set (val_loc p) v)
  | terminates_free : forall s p,
      Fmap.indom s p ->
      terminates s (val_free (val_loc p)).

Section OmnibitTerminates.
Hint Constructors big omnibig.

(** We prove that [omnibig s t Q] is equivalent to the conjunction of
    [terminates s t] and to a partial correctness result asserting that
    if an evaluation of [t/s] terminates on some result then this result
    satisfies [Q]. (omni-big-step-iff-terminates-and-correct) *)

Lemma omnibig_iff_terminates_and_post : forall s t Q,
  omnibig s t Q <-> (terminates s t /\ (forall v s', big s t s' v -> Q v s')).
Proof using.
  iff M.
  { split.
    { induction M; try solve [constructors*].
      { rename M into M1, H into M2, IHM into IH1, H0 into IH2.
        constructors.
        { applys IH1. }
        { introv R. applys IH2. applys omnibig_and_big_inv M1 R. } } }
    { introv R. applys omnibig_and_big_inv M R. } }
  { destruct M as (HS&HQ). gen Q.
    induction HS; try solve [constructors*].
    { rename HS into M1, H into M2, IHHS into IH1, H0 into IH2.
      intros Q HQ.
      applys omnibig_let (fun v1 s2 => big s t1 s2 v1).
      { applys* IH1. }
      { introv R1. applys IH2.
        { applys R1. }
        { introv R2. applys* HQ. } } } }
Qed.

(** We prove that [terminates s t] is equivalent to [omnibig s t Any]. *)

Lemma terminates_iff_omnibig_any : forall s t,
  terminates s t <-> omnibig s t Any.
Proof using.
  intros. rewrite* (omnibig_iff_terminates_and_post s t Any).
Qed.

End OmnibitTerminates.

(** We can reformulate the right to left direction of
    [terminates_iff_omnibig_any] by stating that if [omnibig]
    holds for some [Q], then [terminates s t] holds. *)

Lemma terminates_of_omnibig : forall s t Q,
  omnibig s t Q ->
  terminates s t.
Proof using.
  introv M. rewrite terminates_iff_omnibig_any.
  applys omnibig_conseq M. intros s' v. auto.
Qed.

(** We can reformulate [omnibig_strongestpost_of_omnibig]
    with [terminates] as premise. (strongest-post-correct) *)

Lemma omnibig_strongestpost_of_terminates : forall s t,
  terminates s t ->
  omnibig s t (strongestpost s t).
Proof using.
  introv M. rewrite terminates_iff_omnibig_any in M.
  applys omnibig_strongestpost_of_omnibig M.
Qed.


(* ########################################################### *)
(* ########################################################### *)
(* ########################################################### *)
(** * Inductive PRECISE-Omni-Big-Step  *)

(* ########################################################### *)
(** ** Definition of precise Omni-semantics *)

(** Judgment [pomnibig s t Q] with precise outcome *)

Inductive pomnibig : state -> trm -> (val->state->Prop) -> Prop :=
  | pomnibig_val : forall s v Q,
      Q = exactly v s ->
      pomnibig s (trm_val v) Q
  | pomnibig_fix : forall s f x t1 Q,
      Q = exactly (val_fix f x t1) s ->
      pomnibig s (trm_fix f x t1) Q
  | pomnibig_app_fix : forall s v1 v2 f x t1 Q,
      v1 = val_fix f x t1 ->
      pomnibig s (subst x v2 (subst f v1 t1)) Q ->
      pomnibig s (trm_app v1 v2) Q
  | pomnibig_let : forall Q1 s x t1 t2 Q (Qof:val->state->val->hprop),
      pomnibig s t1 Q1 ->
      (forall v1 s2, Q1 v1 s2 -> pomnibig s2 (subst x v1 t2) (Qof v1 s2)) ->
      Q = (fun v0 s0 => exists s1 v1, Q1 v1 s1 /\ Qof v1 s1 v0 s0) ->
      pomnibig s (trm_let x t1 t2) Q
  | pomnibig_if : forall s b t1 t2 Q,
      pomnibig s (if b then t1 else t2) Q ->
      pomnibig s (trm_if (val_bool b) t1 t2) Q
  | pomnibig_add : forall s n1 n2 Q,
      Q = exactly (val_int (n1 + n2)) s ->
      pomnibig s (val_add (val_int n1) (val_int n2)) Q
  | pomnibig_rand : forall s n Q (Qof:int->val->hprop),
      n > 0 ->
      (forall n1, 0 <= n1 < n -> Qof n1 = exactly (val_int n1) s) ->
      Q = (fun v0 s0 => exists n1, 0 <= n1 < n /\ Qof n1 v0 s0) ->
      pomnibig s (val_rand (val_int n)) Q
  | pomnibig_ref : forall s v Q (Qof:loc->val->hprop),
      (forall p, ~ Fmap.indom s p ->
         Qof p = exactly (val_loc p) (Fmap.update s p v)) ->
      Q = (fun v0 s0 => exists p,  ~ Fmap.indom s p /\ Qof p v0 s0) ->
      pomnibig s (val_ref v) Q
  | pomnibig_get : forall s p Q,
      Fmap.indom s p ->
      Q = exactly (Fmap.read s p) s ->
      pomnibig s (val_get (val_loc p)) Q
  | pomnibig_set : forall s p v Q,
      Fmap.indom s p ->
      Q = exactly val_unit (Fmap.update s p v) ->
      pomnibig s (val_set (val_loc p) v) Q
  | pomnibig_free : forall s p Q,
      Fmap.indom s p ->
      Q = exactly val_unit (Fmap.remove s p) ->
      pomnibig s (val_free (val_loc p)) Q.


(* ########################################################### *)
(** ** Properties of precise Omni-semantics *)

(** [pomnibig s t Q] relates a program with exactly the set [Q]
    that corresponds exactly to the set of possible results,
    no more, no less. *)

Lemma pomnibig_characterizes_strongestpost : forall s t Q,
  pomnibig s t Q ->
  Q = strongestpost s t.
Proof using.
  introv M. induction M.
  { subst. extens; intros v' s'. unfold exactly, strongestpost.
    iff (->&->) M1. { constructors*. } { inverts* M1. } }
  { subst. extens; intros v' s'. unfold exactly, strongestpost.
    iff (->&->) M1. { constructors*. } { inverts* M1. } }
  { subst. extens; intros v' s'. unfold strongestpost. iff M1.
    { constructors*. } { inverts M1 as E. inverts* E. } }
  { rename H0 into K.
    subst. extens; intros v' s'. unfolds strongestpost. iff (s1&v1&R1&M1) M1.
    { rewrite* K in M1. constructors*. }
    { inverts M1 as R1 R2. exists. split*. rewrite* K. } }
  { subst. extens; intros v' s'. unfold strongestpost. iff M1.
    { applys* big_if. } { inverts* M1. } }
  { subst. extens; intros v' s'. unfold exactly, strongestpost.
    iff (->&->) M1. { constructors*. } { inverts* M1. } }
  { rename H0 into K.
    subst. extens; intros v' s'. unfold exactly, strongestpost.
    iff (n1&E1&R1) M1.
    { rewrite* K in R1. inverts R1. applys* big_rand. }
    { inverts M1; tryfalse. exists. split*. rewrite* K. applys exactly_intro. } }
  { rename H into K.
    subst. extens; intros v' s'. unfold exactly, strongestpost.
    iff (p&E1&R1) M1.
    { rewrite* K in R1. inverts R1. applys* big_ref. }
    { inverts M1; tryfalse. exists. split*. rewrite* K. applys exactly_intro. } }
  { subst. extens; intros v' s'. unfold exactly, strongestpost.
    iff (->&->) M1. { constructors*. } { inverts M1; tryfalse. auto. } }
  { subst. extens; intros v' s'. unfold exactly, strongestpost.
    iff (->&->) M1. { constructors*. } { inverts M1; tryfalse. auto. } }
  { subst. extens; intros v' s'. unfold exactly, strongestpost.
    iff (->&->) M1. { constructors*. } { inverts M1; tryfalse. auto. } }
Qed.

(** Assume [pomnibig s t Q] to hold. Prove that the postcondition [Q] holds of
    any output that [(s,t)] may evaluate to according to the judgment [big]. *)

Lemma pomnibig_and_big_inv : forall s t v s' Q,
  pomnibig s t Q ->
  big s t s' v ->
  Q v s'.
Proof using.
  introv M R. lets E: pomnibig_characterizes_strongestpost M.
  rewrite E. apply R.
Qed.

(** If [pomnibig s t Q] holds, then there exists an output
    [(s',v)] that [(s,t)] may evaluate to according to the judgment
    [big], and that this output satisfies [Q]. *)

Lemma pomnibig_to_one_big : forall s t Q,
  pomnibig s t Q ->
  exists s' v, big s t s' v /\ Q v s'.
Proof using.
  introv M. cuts (s'&v&R): (exists s' v, big s t s' v).
  { exists. split. { applys R. } { applys pomnibig_and_big_inv M R. } }
  induction M.
  { exists. applys big_val. }
  { exists. applys big_fix. }
  { rename IHM into IHM1, M into M1.
    forwards (s'&v&R): IHM1.
    exists. applys* big_app_fix R. }
  { rename IHM into IHM1, H0 into IHM2, M into M1, H into M2.
    forwards (s1'&v1&R1): IHM1.
    lets HQ1: pomnibig_and_big_inv M1 R1.
    forwards (s'&v&R2): IHM2 HQ1.
    exists. applys big_let R1 R2. }
  { forwards (s'&v&R1): IHM.
    exists. applys* big_if R1. }
  { exists. applys* big_add. }
  { rename H0 into N.
    exists. applys* big_rand 0. math. }
  { forwards* (p&D&N): (exists_fresh null s).
    exists. applys* big_ref. }
  { exists. applys* big_get. }
  { exists. applys* big_set. }
  { exists. applys* big_free. }
Qed.

(** Auxilary lemma for the next proof *)
Lemma strongestpost_eq_exactly : forall s t v,
  (forall v' s', big s t s' v' <-> v' = v /\ s' = s) ->
  strongestpost s t = exactly v s.
Proof using.
  introv M. unfold strongestpost, exactly. extens*.
Qed.

(** If a term is safe according to [omnibig], which overapproximates
    the set of results, then [pcpbig], which is precise, holds. *)
Lemma pomnibig_of_omnibig : forall s t Q,
  omnibig s t Q ->
  pomnibig s t (strongestpost s t).
Proof using.
  hint exactly_intro. introv M. induction M.
  { constructors. applys strongestpost_eq_exactly.
   intros. iff R (->&->). { inverts* R. } { constructors*. } }
  { constructors. applys strongestpost_eq_exactly.
    intros. iff R (->&->). { inverts* R. } { constructors*. } }
  { applys* pomnibig_app_fix. applys_eq IHM. extens.
    unfold strongestpost. intros v' s'. iff R.
    { subst. inverts R as E. inverts* E. } { applys* big_app_fix. } }
  { rename IHM into IH1, H0 into IH2.
    applys pomnibig_let (strongestpost s t1)
     (fun v1 s2 => strongestpost s2 (subst x v1 t2)) IH1.
    { intros v1 s2 M1. applys IH2. applys* omnibig_inv_strongestpost_qimpl M. }
    { unfold strongestpost. extens. intros v' s'. iff R (?&?&?&?).
      { inverts* R. } { constructors*. } } }
  { constructors. applys_eq IHM. extens. unfold strongestpost.
    intros v' s'. iff R. { inverts* R. } { constructors*. } }
  { constructors. applys strongestpost_eq_exactly.
    intros. iff R (->&->). { inverts* R. } { constructors*. } }
  { applys* pomnibig_rand. simpl. unfold strongestpost. extens.
    intros v' s'. iff R (?&?&(->&->)). { inverts R; tryfalse. exists* n1. }
    { applys* big_rand. } }
  { applys* pomnibig_ref. simpl. unfold strongestpost. extens.
    intros v' s'. iff R (?&?&(->&->)). { inverts R; tryfalse. exists* p. }
    { applys* big_ref. } }
  { constructors*. applys strongestpost_eq_exactly.
    intros. iff R (->&->). { inverts R; tryfalse; eauto. } { constructors*. } }
  { constructors*. unfold strongestpost, exactly. extens*.
    intros v' s'. iff R (->&->). { inverts R; tryfalse; eauto. } { constructors*. } }
  { constructors*. unfold strongestpost, exactly. extens*.
    intros v' s'. iff R (->&->). { inverts R; tryfalse; eauto. } { constructors*. } }
Qed.

(** See also [omnibig_inv_strongestpost_qimpl], which asserts that
    if [omnibig s t Q], then [Q] contains [strongestpost s t]. *)



(* ########################################################### *)
(* ########################################################### *)
(* ########################################################### *)
(** * Omni-Small-Step *)

Implicit Types P : state->trm->Prop.


(* ########################################################### *)
(** ** Definition of Omni-Small-Step *)

(** [omnismall s t P] asserts that a configuration [(s,t)] can
    take a step, and that for any step it takes it reaches a
    configuration that satisfies [P]. *)

Inductive omnismall : state -> trm -> (state->trm->Prop) -> Prop :=
  (* under context *)
  | omnismall_let_ctx : forall s1 x t1 t2 P1 P,
      omnismall s1 t1 P1 ->
      (forall s2 t1', P1 s2 t1' -> P s2 (trm_let x t1' t2)) ->
      omnismall s1 (trm_let x t1 t2) P
  (* term constructs *)
  | omnismall_fix : forall s f x t1 P,
      P s (val_fix f x t1) ->
      omnismall s (trm_fix f x t1) P
  | omnismall_app_fix : forall s v1 v2 f x t1 P,
      v1 = val_fix f x t1 ->
      P s (subst x v2 (subst f v1 t1)) ->
      omnismall s (trm_app v1 v2) P
  | omnismall_if : forall s b t1 t2 P,
      P s (if b then t1 else t2) ->
      omnismall s (trm_if (val_bool b) t1 t2) P
  | omnismall_let : forall s x t2 v1 P,
      P s (subst x v1 t2) ->
      omnismall s (trm_let x v1 t2) P
  (* primitive operations *)
  | omnismall_add : forall s n1 n2 P,
      P s (val_int (n1 + n2)) ->
      omnismall s (val_add (val_int n1) (val_int n2)) P
  | omnismall_rand : forall s n P,
      n > 0 ->
      (forall n1, 0 <= n1 < n -> P s n1) -> (* non-det *)
      omnismall s (val_rand (val_int n)) P
  | omnismall_ref : forall s v P,
      (forall p, ~ Fmap.indom s p ->
         P (Fmap.update s p v) (val_loc p)) -> (* non-det *)
      omnismall s (val_ref v) P
  | omnismall_get : forall s p P,
      Fmap.indom s p ->
      P s (Fmap.read s p) ->
      omnismall s (val_get (val_loc p)) P
  | omnismall_set : forall s p v P,
      Fmap.indom s p ->
      P (Fmap.update s p v) val_unit ->
      omnismall s (val_set (val_loc p) v) P
  | omnismall_free : forall s p P,
      Fmap.indom s p ->
      P (Fmap.remove s p) val_unit ->
      omnismall s (val_free (val_loc p)) P.


(* ########################################################### *)
(** ** Equivalence of Omni-Small-Step with Standard Small Step *)

Section OmnismallEquiv.
Hint Constructors step.

(* Characterization lemma (omni-small-step-iff-progress-and-correct). *)

Lemma omnismall_iff_step_st : forall s t P,
       omnismall s t P
  <-> (   (exists s' t', step s t s' t')
       /\ (forall s' t', step s t s' t' -> P s' t')).
Proof using.
  iff M.
  { induction M.
    { forwards (R&M1): IHM. split.
      { destruct R as (s'&t'&S). exists. applys step_let_ctx S. }
      { intros s' t' S. inverts S as.
        { rename H into IHM1. introv S1. lets K1: M1 S1. applys IHM1 K1. }
        { destruct R as (s''&t''&S'). false* step_val_inv. } } }
    { split*. { intros s' t' S. inverts S as; try false_invert. { auto. } } }
    { split*. { intros s' t' S. inverts S as; try false_invert.
                { inverts* TEMP. } } }
    { split*. { intros s' t' S. inverts S as; try false_invert. { auto. } } }
    { split*. { intros s' t' S. inverts S as; try false_invert. { auto. } } }
    { split*. { intros s' t' S. inverts S as; try false_invert. { auto. } } }
    { split.
       { exists. applys step_rand 0. math. }
       { intros s' t' S. inverts S as; try false_invert.
         { intros R. inverts R. auto. } } }
    { split.
       { forwards~ (p&F&N): (exists_fresh null s).
         exists. applys* step_ref p. }
       { intros s' t' S. inverts S as; try false_invert. { auto. } } }
    { split*. { intros s' t' S. inverts S as; try false_invert. { auto. } } }
    { split*. { intros s' t' S. inverts S as; try false_invert. { auto. } } }
    { split*. { intros s' t' S. inverts S as; try false_invert. { auto. } } } }
  { destruct M as (R&M). destruct R as (s'&t'&S).
    gen P. induction S; intros.
    { rename S into S1. applys omnismall_let_ctx (step s1 t1).
      { applys IHS. auto. }
      { intros s' t' S1'. applys M. applys step_let_ctx S1'. } }
    { applys* omnismall_fix. }
    { applys* omnismall_app_fix. }
    { applys* omnismall_if. }
    { applys* omnismall_let. }
    { applys* omnismall_add. }
    { applys* omnismall_rand. math. }
    { applys* omnismall_ref. }
    { applys* omnismall_get. }
    { applys* omnismall_set. }
    { applys* omnismall_free. } }
Qed.

End OmnismallEquiv.


(* ########################################################### *)
(** ** Eventually Judgment *)

(** [eventually s t P] asserts that all executions starting from [(s,t)]
    eventually reach a configuration [(s',t')] satisfying the predicate [P]. *)

(** Definition of [eventually s t P] in terms of [omnismall]. *)

Inductive eventually : state->trm->(state->trm->Prop)->Prop :=
  | eventually_here : forall s t P,
      P s t ->
      eventually s t P
  | eventually_step : forall s t P1 P,
      omnismall s t P1 ->
      (forall s' t', P1 s' t' -> eventually s' t' P) ->
      eventually s t P.

(** Alternative definition of [eventually], with respect to standard small-step.
    (eventually-step-using-standard-small-step) *)

Inductive eventually' : state->trm->(state->trm->Prop)->Prop :=
  | eventually'_here : forall s t P,
      P s t ->
      eventually' s t P
  | eventually'_step : forall s t P,
      (exists s' t', step s t s' t') ->
      (forall s' t', step s t s' t' -> eventually' s' t' P) ->
      eventually' s t P.

(** [eventually] is equivalent to [eventually']. *)

Lemma eventually_eq_eventually' :
  eventually = eventually'.
Proof using.
  extens. intros s t P. iff M.
  { induction M.
    { applys* eventually'_here. }
    { rename H into M1, H0 into M2, H1 into IHM2.
      rewrite omnismall_iff_step_st in M1. destruct M1 as (R&M1).
      applys* eventually'_step R. } }
  { induction M.
    { applys* eventually_here. }
    { rename H into R, H0 into M1, H1 into IHM1.
      applys eventually_step (step s t).
      { rewrite omnismall_iff_step_st. split*. }
      { autos*. } } }
Qed.

(** Chained rule for [eventually_step]. *)

Lemma eventually_step_chained : forall s t P,
  omnismall s t (fun s' t' => eventually s' t' P) ->
  eventually s t P.
Proof using. introv M. applys* eventually_step M. Qed.

(** The cut rule for [eventually]. *)

Lemma eventually_cut : forall s t P1 P2,
  eventually s t P1 ->
  (forall s' t', P1 s' t' -> eventually s' t' P2) ->
  eventually s t P2.
Proof using.
  introv M HK. induction M.
  { rename H into M1. applys HK M1. }
  { rename H into R, H0 into M1, H1 into IH1.
    applys eventually_step R. intros s' t' S.
    applys IH1 S HK. }
Qed.

(** Chained rule for [eventually_cut]. *)

Lemma eventually_cut_chained : forall s t P,
  eventually s t (fun s' t' => eventually s' t' P) ->
  eventually s t P.
Proof using. introv M. applys* eventually_cut M. Qed.


(* ########################################################### *)
(** ** Inductive Small-Step Judgment *)

(** The judgment [bigsmall] introduced in this section helps
    carrying out several proofs of equivalence---see the next
    section. *)

(** Viewing postconditions as predicates over configurations *)

Definition pred_of_post (Q:val->hprop) : state->trm->Prop :=
  fun s' t' => exists v', t' = trm_val v' /\ Q v' s'.

(** We introduce the inductive small-step judgment, as it is useful
    to relate Omni-Small-Step with Omni-Big-Step.

    [bigsmall s t Q] asserts that executions of [(t,s)] terminate,
    and the results satisfy the postcondition [Q]. *)

Inductive bigsmall : state->trm->(val->hprop)->Prop :=
  | bigsmall_val : forall s v Q,
      Q v s ->
      bigsmall s v Q
  | bigsmall_step : forall s t Q,
      (exists s' t', step s t s' t') ->
      (forall s' t', step s t s' t' -> bigsmall s' t' Q) ->
      bigsmall s t Q.

(** [bigsmall] is equivalent to [eventually]. *)

Lemma bigsmall_iff_eventually : forall s t Q,
  bigsmall s t Q <-> eventually s t (pred_of_post Q).
Proof using.
  rewrite eventually_eq_eventually'.
  iff M.
  { induction M.
    { applys eventually'_here. hnf. exists*. }
    { rename H into R, H0 into M1, H1 into IH1.
      applys eventually'_step R. applys IH1. } }
  { gen_eq C: (pred_of_post Q). induction M; intros; subst.
    { destruct H as (v'&->&Hv'). applys bigsmall_val Hv'. }
    { rename H into R, H0 into M1, H1 into IH1.
      applys bigsmall_step R.
      intros s' t' S. applys* IH1 S. } }
Qed.


(* ########################################################### *)
(** ** Equivalence Between Eventually and Omni-Big-Step Semantics *)

(** We start by establishing the equivalence between [bigsmall] and [omnibig].
    We focus first on the direction from [bigsmall] to [omnibig]. *)

(** We begin with a key lemma: if a configuration [(s1,t1)] takes a step to
    [(s2,t2)], then this first configuration admits the same postconditions
    as the second configuration. *)

Lemma omnibig_of_step_and_omnibig : forall s1 t1 Q,
  (exists s2 t2, step s1 t1 s2 t2) ->
  (forall s2 t2, step s1 t1 s2 t2 -> omnibig s2 t2 Q) ->
  omnibig s1 t1 Q.
Proof using.
  introv (s2&t2&R1) RS. gen Q. induction R1; intros.
  { applys omnibig_let (fun v1 s2 => omnibig s2 (subst x v1 t2) Q).
    { applys IHR1. intros s1b t1b Kb.
      forwards M: RS. { applys step_let_ctx Kb. }
      { inverts M as M1 M2. applys omnibig_conseq M1. applys M2. } }
    { intros v1 s' K. applys K. } }
  { applys omnibig_fix. forwards M: RS. { applys step_fix. }
    { inverts* M. } }
  { applys* omnibig_app_fix. forwards M: RS. { applys* step_app_fix. }
    { applys M. } }
  { applys* omnibig_if. forwards M: RS. { applys* step_if. } { applys M. } }
  { applys omnibig_let (fun v' s' => v' = v1 /\ s' = s).
    { applys* omnibig_val. }
    { intros ? ? (->&->). forwards M: RS. { applys* step_let. }
      { applys M. } } }
  { applys* omnibig_add. forwards M: RS. { applys* step_add. }
    { inverts* M. } }
  { applys* omnibig_rand.
    { math. }
    { intros n2 N2. forwards M: RS. { applys* step_rand n2. }
      { inverts* M. } } }
  { applys omnibig_ref. intros p' D.
    forwards M: RS p'. { applys* step_ref. } { inverts* M. } }
  { applys* omnibig_get. forwards M: RS. { applys* step_get. }
    { inverts* M. } }
  { applys* omnibig_set. forwards M: RS. { applys* step_set. }
    { inverts* M. } }
  { applys* omnibig_free. forwards M: RS. { applys* step_free. }
    { inverts* M. } }
Qed.

(** By exploiting the above lemma over all the steps of an execution, we obtain
    the fact that [bigsmall] implies [omnibig]. *)

Lemma omnibig_of_bigsmall : forall s t Q,
  bigsmall s t Q ->
  omnibig s t Q.
Proof using.
  introv M. induction M.
  { applys* omnibig_val. }
  { applys* omnibig_of_step_and_omnibig. }
Qed.

(** Let's now turn to the second direction, from [omnibig] to [bigsmall].
    We first need an introduction lemma for let-bindings for [bigsmall]. *)

Lemma bigsmall_let : forall s x t1 t2 Q1 Q,
  bigsmall s t1 Q1 ->
  (forall s1 v1, Q1 v1 s1 -> bigsmall s1 (subst x v1 t2) Q) ->
  bigsmall s (trm_let x t1 t2) Q.
Proof using.
  introv M1 M2. gen_eq Q1': Q1.
  induction M1; intros; subst.
  { apply bigsmall_step.
    { do 2 esplit. applys step_let. }
    { introv R. inverts R as R'. { inverts R'. } applys* M2. } }
  { rename t into t1, H1 into IH.
    destruct H as (s'&t'&RE). applys bigsmall_step.
    { do 2 esplit. constructors. applys RE. }
    { introv R. inverts R as R'.
      { applys* IH R' M2. }
      { false. inverts RE. } } }
Qed.

(** The proof is carried out by induction on the big-step relation. *)

Lemma bigsmall_of_omnibig : forall s t Q,
  omnibig s t Q ->
  bigsmall s t Q.
Proof using.
  introv M. induction M.
  { applys* bigsmall_val. }
  { applys bigsmall_step.
    { do 2 esplit. applys step_fix. }
    { introv K. inverts K. applys* bigsmall_val. } }
  { rename H into E. applys bigsmall_step.
    { do 2 esplit. applys* step_app_fix. }
    { introv K. invert K; try solve [ intros; false ].
      introv -> -> -> -> -> R. inverts E. applys IHM. } }
  { rename M into M1, H into M2, IHM into IHM1, H0 into IHM2.
    tests C: (exists v1, t1 = trm_val v1).
    { destruct C as (v1&->). applys bigsmall_step.
      { do 2 esplit. applys* step_let. }
      { introv K. inverts K as K1 K2.
        { inverts K1. }
        { inverts IHM1 as K3 K4.
          { applys* IHM2. }
          { destruct K3 as (?&?&K5). inverts K5. } } } }
    { applys bigsmall_step.
      { inverts IHM1 as K3 K4.
        { false* C. }
        { destruct K3 as (?&?&K5). do 2 esplit. applys* step_let_ctx. } }
      { introv K. inverts K as K'; [|false* C].
        inverts IHM1 as K5 K6; [false* C|].
        specializes K6 K'. applys* bigsmall_let. } } }
  { applys bigsmall_step.
    { do 2 esplit. applys* step_if. }
    { introv K. inverts K. applys IHM. } }
  { applys bigsmall_step.
    { do 2 esplit. applys* step_add. }
    { introv K. inverts K. applys* bigsmall_val. } }
  { applys bigsmall_step.
    { do 2 esplit. applys* step_rand 0. math. }
    { introv K. inverts K; tryfalse. applys* bigsmall_val. } }
  { applys bigsmall_step.
    { forwards~ (p&D&N): (exists_fresh null s).
      do 2 esplit. applys* step_ref. }
    { introv K. inverts K; tryfalse. applys* bigsmall_val. } }
  { applys bigsmall_step.
    { do 2 esplit. applys* step_get. }
    { introv K. inverts K; tryfalse. applys* bigsmall_val. } }
  { applys bigsmall_step.
    { do 2 esplit. applys* step_set. }
    { introv K. inverts K; tryfalse. applys* bigsmall_val. } }
  { applys bigsmall_step.
    { do 2 esplit. applys* step_free. }
    { introv K. inverts K; tryfalse. applys* bigsmall_val. } }
Qed.

(** Combining the two directions, we obtain the desired equivalence between
    omni-big-step and inductive-small-step. *)

Lemma bigsmall_eq_omnibig :
  bigsmall = omnibig.
Proof using.
  extens. intros s t Q. iff M.
  { applys* omnibig_of_bigsmall. }
  { applys* bigsmall_of_omnibig. }
Qed.

(** The final equivalence follows. *)

Lemma omnibig_iff_eventually : forall s t Q,
  omnibig s t Q <-> eventually s t (pred_of_post Q).
Proof using.
  intros. rewrite <- bigsmall_eq_omnibig.
  rewrite* bigsmall_iff_eventually.
Qed.


(* ########################################################### *)
(* ########################################################### *)
(* ########################################################### *)
(** * Coinductive Omni-big-step *)

(* ########################################################### *)
(** ** Definition of Coinductive Omni-Big-Step *)

(** The definition of the predicate [coomnibig] is obtained by taking the
    constructors from the inductive definition of [steps], and considering
    the coinductive interpretation of these constructors. The coinductive
    interpretation allows for infinite derivation. It thereby introduces the
    possibility of diverging executions. Importantly, the predicate [coomnibig]
    still disallows executions that get stuck. *)

CoInductive coomnibig : state -> trm -> (val->state->Prop) -> Prop :=
  | coomnibig_val : forall s v Q,
      Q v s ->
      coomnibig s (trm_val v) Q
  | coomnibig_fix : forall s f x t1 Q,
      Q (val_fix f x t1) s ->
      coomnibig s (trm_fix f x t1) Q
  | coomnibig_app_fix : forall s1 v1 v2 f x t1 Q,
      v1 = val_fix f x t1 ->
      coomnibig s1 (subst x v2 (subst f v1 t1)) Q ->
      coomnibig s1 (trm_app v1 v2) Q
  | coomnibig_let : forall Q1 s1 x t1 t2 Q,
      coomnibig s1 t1 Q1 ->
      (forall v1 s2, Q1 v1 s2 -> coomnibig s2 (subst x v1 t2) Q) ->
      coomnibig s1 (trm_let x t1 t2) Q
  | coomnibig_if : forall s1 b t1 t2 Q,
      coomnibig s1 (if b then t1 else t2) Q ->
      coomnibig s1 (trm_if (val_bool b) t1 t2) Q
  | coomnibig_add : forall s n1 n2 Q,
      Q (val_int (n1 + n2)) s ->
      coomnibig s (val_add (val_int n1) (val_int n2)) Q
  | coomnibig_rand : forall s n Q,
      n > 0 ->
      (forall n1, 0 <= n1 < n -> Q n1 s) ->
      coomnibig s (val_rand (val_int n)) Q
  | coomnibig_ref : forall s v Q,
      (forall p, ~ Fmap.indom s p ->
         Q (val_loc p) (Fmap.update s p v)) ->
      coomnibig s (val_ref v) Q
  | coomnibig_get : forall s p Q,
      Fmap.indom s p ->
      Q (Fmap.read s p) s ->
      coomnibig s (val_get (val_loc p)) Q
  | coomnibig_set : forall s p v Q,
      Fmap.indom s p ->
      Q val_unit (Fmap.update s p v) ->
      coomnibig s (val_set (val_loc p) v) Q
  | coomnibig_free : forall s p Q,
      Fmap.indom s p ->
      Q val_unit (Fmap.remove s p) ->
      coomnibig s (val_free (val_loc p)) Q.

(** Covariance property. *)

Lemma coomnibig_conseq : forall s t Q1 Q2,
  coomnibig s t Q1 ->
  Q1 ===> Q2 ->
  coomnibig s t Q2.
Proof using.
  introv M W. unfolds qimpl, himpl.
  gen s t Q1 Q2. cofix IH. intros. inverts M; try solve [ constructors* ].
Qed.

(** Assume [coomnibig s t Q] to hold. We prove that the postcondition [Q] holds of
    any output that [(s,t)] may evaluate to according to the judgment [big]. *)

Lemma coomnibig_inv_eval : forall s t v s' Q,
  coomnibig s t Q ->
  big s t s' v ->
  Q v s'.
Proof using.
  introv M R. gen Q. induction R; intros;
   try solve [inverts* M; tryfalse].
  { inverts M; tryfalse. inverts TEMP.
    applys* IHR Q. }
  { inverts M as M1 M2. applys IHR2 M2. applys IHR1 M1. }
  { inverts M. applys* IHR Q. }
Qed.


(* ########################################################### *)
(** ** From Total to Partial Correctness *)

(** Total correctness is a stronger property than partial correctness:
    [omnibig s t Q] entails [coomnibig s t Q]. *)

Lemma coomnibig_of_omnibig : forall s t Q,
  omnibig s t Q ->
  coomnibig s t Q.
Proof using.
  introv M. induction M; try solve [ constructors* ].
Qed.


(* ########################################################### *)
(** ** Interpretation of Coinductive Omni-Big-Step w.r.t. [safe] *)

(** The predicate [safe s t] asserts that an execution of the configuration
    [t/s] can never get stuck. Its definition is a simplified version of
    [coomnibig] where all occurences of [Q] are removed. In the rule for let-
    bindings, namely [safe_let], the quantification over the configuration
    [v1/s2] is done by refering to the big-step judgment [eval]. *)

CoInductive safe : state -> trm -> Prop :=
  | safe_val : forall s v,
      safe s (trm_val v)
  | safe_fix : forall s f x t1,
      safe s (trm_fix f x t1)
  | safe_app_fix : forall s v1 v2 f x t1,
      v1 = val_fix f x t1 ->
      safe s (subst x v2 (subst f v1 t1)) ->
      safe s (trm_app v1 v2)
  | safe_let : forall s x t1 t2,
      safe s t1 ->
      (forall v1 s2, big s t1 s2 v1 -> safe s2 (subst x v1 t2)) ->
      safe s (trm_let x t1 t2)
  | safe_if : forall s b t1 t2,
      safe s (if b then t1 else t2) ->
      safe s (trm_if (val_bool b) t1 t2)
  | safes_add : forall s n1 n2,
      safe s (val_add (val_int n1) (val_int n2))
  | safe_rand : forall s n,
      n > 0 ->
      safe s (val_rand (val_int n))
  | safe_ref : forall s v,
      safe s (val_ref v)
  | safe_get : forall s p,
      Fmap.indom s p ->
      safe s (val_get (val_loc p))
  | safe_set : forall s p v,
      Fmap.indom s p ->
      safe s (val_set (val_loc p) v)
  | safe_free : forall s p,
      Fmap.indom s p ->
      safe s (val_free (val_loc p)).

Section EvalnpSafe.
Hint Constructors big coomnibig.

(** We prove that [coomnibig s t Q] is equivalent to the conjunction of [safe s t]
    and to a partial correctness result asserting that if an evaluation of
    [t/s] terminates on a particular result, then this result satisfies [Q].
    omni-cobig-iff-safe-and-correct *)

Lemma coomnibig_iff_safe_and_post : forall s t Q,
  coomnibig s t Q <-> (safe s t /\ (forall v s', big s t s' v -> Q v s')).
Proof using.
  iff M.
  { split.
    { gen s t Q. cofix IH. introv M. inverts M; try solve [constructors*].
      { rename H into M1, H0 into M2. constructors.
        { applys IH M1. }
        { introv R. applys IH M2. applys coomnibig_inv_eval M1 R. } } }
    { introv R. applys coomnibig_inv_eval M R. } }
  { destruct M as (HS&HQ). gen s t Q. cofix IH. intros. inverts HS;
     try solve [constructors*].
    { rename H into M1, H0 into M2.
      applys coomnibig_let (fun v1 s2 => big s t1 s2 v1).
      { applys* IH M1. }
      { introv R1. applys IH M2.
        { applys R1. }
        { introv R2. applys* HQ. } } } }
Qed.

(** We prove that [safe s t] is equivalent to [coomnibig s t Any]. *)

Lemma safe_iff_coomnibig_any : forall s t,
  safe s t <-> coomnibig s t Any.
Proof using.
  intros. rewrite* (coomnibig_iff_safe_and_post s t Any).
Qed.

End EvalnpSafe.


(* ########################################################### *)
(** ** Small-Step Characterization of Partial Correctness *)

(** [partial s t Q] asserts that the executions of [(t,s)] are safe
    and that if they terminate they satisfy the postcondition [Q]. *)

Definition partial (s1:state) (t1:trm) (Q:val->state->Prop) : Prop :=
  forall s2 t2, steps s1 t1 s2 t2 ->
       (exists v2, t2 = trm_val v2 /\ Q v2 s2)
    \/ (exists s3 t3, step s2 t2 s3 t3).

(** Inversion lemma *)

Lemma partial_val_inv : forall s v Q,
  partial s v Q ->
  Q v s.
Proof using.
  introv M. forwards N: M s v. { applys steps_refl. }
  destruct N as [(v1&E1&P1)|(s2&t2&R)].
  { inverts* E1. }
  { inverts R; false_invert. } (* patch one line *)
Qed.

(** Preservation property for [partial]. *)

Lemma partial_inv_step : forall s1 t1 s2 t2 Q,
  partial s1 t1 Q ->
  step s1 t1 s2 t2 ->
  partial s2 t2 Q.
Proof using.
  introv M R1. intros s3 t3 R2. applys M s3 t3.
  { applys steps_step R1. applys R2. }
Qed.

(** Progress property for [partial]. *)

Lemma partial_inv : forall s1 t1 Q,
  partial s1 t1 Q ->
     (exists v1, t1 = trm_val v1 /\ Q v1 s1)
  \/ (exists s2 t2, step s1 t1 s2 t2 /\ partial s2 t2 Q).
Proof using.
  introv M. forwards N: M s1 t1. { applys steps_refl. }
  destruct N as [(v1&E1&P1)|(s2&t2&R)].
  { left. exists* v1. }
  { right. exists s2 t2. split~. { applys* partial_inv_step. } }
Qed.

(** An alternative statement of the previous lemma. *)

Lemma partial_not_val_inv : forall s1 t1 Q,
  partial s1 t1 Q ->
  (forall v, t1 <> trm_val v) ->
  exists s2 t2, step s1 t1 s2 t2 /\ partial s2 t2 Q.
Proof using.
  introv M N. destruct (partial_inv M) as [(v1&E1&P1)|R].
  { false* N. }
  { auto. }
Qed.


(* ########################################################### *)
(** ** CoInductive Small-Step Judgment *)

(** Here again, we introduce an auxiliary judgment for the sake of
    the equivalence proof in the next section. *)

(** [cobigsmall s t Q] asserts that all executions of [(s,t)] are
    safe, and if they terminate then they satisfy the postcondition [Q]. *)

CoInductive cobigsmall : state->trm->(val->hprop)->Prop :=
  | cobigsmall_val : forall s v Q,
      Q v s ->
      cobigsmall s v Q
  | cobigsmall_step : forall s t Q,
      (exists s' t', step s t s' t') ->
      (forall s' t', step s t s' t' -> cobigsmall s' t' Q) ->
      cobigsmall s t Q.

(** [cobigsmall] patches the standard definition of partial correctness
    with respect to standard small-step. *)

Lemma cobigsmall_eq_partial :
  cobigsmall = partial.
Proof using.
  extens. intros s t Q. iff M.
  { introv R. gen M. induction R; intros.
    { inverts M as M1 M2.
      { left*. }
      { right. applys M1. } }
    { rename H into R1, R into R2. inverts M as M1 M2.
      { false. inverts R1; false_invert. }
      { applys IHR. applys M2 R1. } } }
  { gen s t Q. cofix IH. intros.
    tests C: (exists v, t = trm_val v).
    { destruct C as (v&->). applys cobigsmall_val. applys partial_val_inv M. }
    { lets (s2&t2&R2&RS): partial_not_val_inv M C.
      applys cobigsmall_step.
      { exists* s2 t2. }
      { intros s' t' R'. applys IH. applys partial_inv_step M R'. } } }
Qed.


(* ########################################################### *)
(** ** Equivalence Between CoInductive Big-Step and Standard Partial Correctness *)

(** The first step of the proof is to formally relate [cobigsmall]
   and [coomnibig], which are both defined coinductively. *)

(** The first direction is from [cobigsmall] to [coomnibig].
    To establish it, we need inversion lemmas for let-bindings. *)

Lemma cobigsmall_let_inv_ctx : forall x s1 t1 t2 Q Q1,
  cobigsmall s1 (trm_let x t1 t2) Q ->
  (fun v s => steps s1 t1 s v) ===> Q1 ->
  cobigsmall s1 t1 Q1.
Proof using.
  cofix IH. introv M WQ. inverts M as M0 M'.
  tests C: (exists v1, t1 = trm_val v1).
  { destruct C as (v1&->). applys cobigsmall_val.
    applys WQ. applys steps_refl. }
  { applys cobigsmall_step.
    { destruct M0 as (s'&t'&S). inverts S as. 2:{ false* C. }
      intros S. exists. applys S. }
    { intros s2 t1' S. applys IH.
      { applys M'. applys step_let_ctx S. }
      { intros v s R. applys WQ. applys steps_step S R. } } }
Qed.

Lemma cobigsmall_let_inv_cont : forall s1 s2 v1 x t1 t2 Q,
  cobigsmall s1 (trm_let x t1 t2) Q ->
  steps s1 t1 s2 v1 ->
  cobigsmall s2 (subst x v1 t2) Q.
Proof using.
  introv M R. gen_eq t1': (trm_val v1). induction R; intros; subst.
  { inverts M as _ M'. applys M'. applys step_let. }
  { rename H into S, R into R', t0 into t1'.
    inverts M as _ M'. applys* IHR. applys M'.
    applys step_let_ctx S. }
Qed.

Lemma cobigsmall_let_inv : forall s1 x t1 t2 Q,
  cobigsmall s1 (trm_let x t1 t2) Q ->
  exists Q1, cobigsmall s1 t1 Q1
          /\ (forall v1 s2, Q1 v1 s2 -> cobigsmall s2 (subst x v1 t2) Q).
Proof using.
  introv M. exists (fun v s => steps s1 t1 s (trm_val v)). split.
  { applys* cobigsmall_let_inv_ctx M. }
  { introv K. applys* cobigsmall_let_inv_cont M K. }
Qed.

(** Using these inversion lemmas, we can prove by induction the first
    direction, from [cobigsmall] to [coomnibig]. *)

Lemma coomnibig_of_cobigsmall : forall s t Q,
  cobigsmall s t Q ->
  coomnibig s t Q.
Proof using.
  cofix IH. introv M. destruct t; try solve [ inverts M; false_invert ].
  { inverts M as.
    { intros R. applys coomnibig_val R. }
    { intros (s'&t'&S) _. inverts S. } }
  { inverts M as.
    { intros (s'&t'&S) _. inverts S. } }
  { rename v into f, v0 into x, t into t1.
    inverts M as (s'&t'&S) M1. inverts S.
    applys coomnibig_fix.
    forwards M1': M1. { applys step_fix. }
    inverts M1' as. { auto. } { intros. false* exists_step_val_inv. } }
  { inverts M as (s'&t'&S) M1. inverts S as.
    { applys coomnibig_app_fix. { reflexivity. } applys IH.
      applys M1. applys* step_app_fix. }
    { applys coomnibig_add.
      forwards M1': M1. { applys step_add. }
      inverts M1' as. { auto. } { intros. false* exists_step_val_inv. } }
    { introv N. applys* coomnibig_rand. { math. }
      intros n1' N1.
      forwards M1': M1. { applys* step_rand n1'. }
      inverts M1' as. { auto. } { intros. false* exists_step_val_inv. } }
    { introv D. applys* coomnibig_ref. intros p' D'.
      forwards M1': M1. { applys* step_ref p'. }
      inverts M1' as. { auto. } { intros. false* exists_step_val_inv. } }
    { introv D. applys* coomnibig_get.
      forwards M1': M1. { applys* step_get. }
      inverts M1' as. { auto. } { intros. false* exists_step_val_inv. } }
    { introv D. applys* coomnibig_set.
      forwards M1': M1. { applys* step_set. }
      inverts M1' as. { auto. } { intros. false* exists_step_val_inv. } }
    { introv D. applys* coomnibig_free.
      forwards M1': M1. { applys* step_free. }
      inverts M1' as. { auto. } { intros. false* exists_step_val_inv. } } }
  { lets (Q1&M1&M2): cobigsmall_let_inv M. applys coomnibig_let.
    { applys IH. applys M1. }
    { introv R. applys IH. applys M2 R. } }
  { inverts M as (s'&t'&S) M1. inverts S. applys coomnibig_if.
    applys IH. applys M1. applys step_if. }
Qed.

(** For the reciprocal direction, we also need one key inversion lemma. It
    is stated below. Its hypothesis is [coomnibig s t Q], and its conclusion
    corresponds to the disjunction of the constructors of the inductive
    definition of [cobigsmall s t Q]. *)

Lemma coomnibig_inv_step : forall s t Q,
  coomnibig s t Q ->
     (exists v, t = trm_val v /\ Q v s)
  \/ ((exists s' t', step s t s' t' /\ coomnibig s' t' Q)
      /\ (forall s' t', step s t s' t' -> coomnibig s' t' Q)).
Proof using.
  introv M. gen s Q. induction t; intros; inverts M as.
  { introv R. left. eauto. }
  { introv R. right. split.
    { exists. split. { applys step_fix. } { applys* coomnibig_val. } }
    { intros s' t' S. inverts S. applys* coomnibig_val. } }
  { introv M1. right. split.
    { exists. split. { applys* step_app_fix. } { auto. } }
    { intros s' t' S. inverts S as E. inverts E. auto. } }
  { introv R. right. split.
    { exists. split. { applys step_add. } { applys* coomnibig_val. } }
    { intros s' t' S. inverts S. applys* coomnibig_val. } }
  { introv N R. right. split.
    { exists. split. { applys* step_rand 0. math. }
       { applys* coomnibig_val. applys R. math. } }
    { intros s' t' S. inverts S; tryfalse. applys* coomnibig_val. } }
  { introv R. right. split.
    { forwards~ (p&D&N): (exists_fresh null s).
      exists. split. { applys* step_ref. } { applys* coomnibig_val. } }
    { intros s' t' S. inverts S; tryfalse. applys* coomnibig_val. } }
  { introv D R. right. split.
    { exists. split. { applys* step_get. } { applys* coomnibig_val. } }
    { intros s' t' S. inverts S; tryfalse. applys* coomnibig_val. } }
  { introv D R. right. split.
    { exists. split. { applys* step_set. } { applys* coomnibig_val. } }
    { intros s' t' S. inverts S; tryfalse. applys* coomnibig_val. } }
  { introv D R. right. split.
    { exists. split. { applys* step_free. } { applys* coomnibig_val. } }
    { intros s' t' S. inverts S; tryfalse. applys* coomnibig_val. } }
  { introv M1 M2. rename v into x. right. split.
    { forwards [(v1&->&HQ1)|((s'&t'&S'&M')&_)]: IHt1 M1.
      { exists. split. { applys step_let. } { applys* M2. } }
      { exists. split.
        { applys step_let_ctx S'. }
        { applys coomnibig_let M'. applys M2. } } }
    { intros s' t' S. inverts S as S.
      { forwards [(v1&->&HQ1)|(_&M3)]: IHt1 M1.
        { inverts S. }
        { specializes M3 S. applys coomnibig_let M3 M2. } }
      { inverts M1 as R. applys* M2. } } }
  { introv M1. right. split.
    { exists. split. { applys step_if. } { auto. } }
    { intros s' t' S. inverts S. auto. } }
Qed.

(** Using this inversion lemma, it is straightforward to derive the
    implication from [coomnibig] to [cobigsmall]. *)

Lemma cobigsmall_of_coeval : forall s t Q,
  coomnibig s t Q ->
  cobigsmall s t Q.
Proof.
  cofix IH. introv M. lets C: coomnibig_inv_step M.
  destruct C as [(v&->&HQ)|((s'&t'&S&M1)&M2)].
  { applys cobigsmall_val HQ. }
  { applys cobigsmall_step.
    { exists. applys S. }
    { intros s1 t1 S'. applys IH. applys M2 S'. } }
Qed.

(** Combining the two directions yields the equality between [cobigsmall] and
    [coomnibig], and that between [partial] and [coomnibig]. *)

Lemma cobigsmall_eq_omnibigp :
   cobigsmall = coomnibig.
Proof using.
  extens. intros s t Q. iff M.
  { applys* coomnibig_of_cobigsmall. }
  { applys* cobigsmall_of_coeval. }
Qed.

(** Finally, we conclude that coinductive Omni-big-step matches the
    standard small-step characterization of partial correctness. *)

Lemma coomnibig_eq_partial :
  coomnibig = partial.
Proof using.
  rewrite <- cobigsmall_eq_partial. rewrite* cobigsmall_eq_omnibigp.
Qed.


(* ########################################################### *)
(** ** Small-Step Characterization of Safety *)

(** Safety corresponds to partial correctness with the always-true
    postcondition. We let [ssafe] denote the corresponding
    judgment. *)

Definition ssafe (s:state) (t:trm) : Prop :=
  partial s t Any.

(** [ssafe] is equivalent to [safe] *)

Lemma ssafe_eq_safe :
  ssafe = safe.
Proof using.
  extens. intros s t. unfold ssafe.
  rewrite <- coomnibig_eq_partial.
  rewrite* safe_iff_coomnibig_any.
Qed.


(* ########################################################### *)
(* ########################################################### *)
(* ########################################################### *)
(** * Characterization of Divergence *)

(** We present several characterization of divergence and prove
    them equivalent. *)


(* ########################################################### *)
(** ** Omni-Big-Step Divergence *)

(** The judgment [coomnidiv s t] asserts that every possible
    execution of the program [(s,t)] diverges. It is defined
    in terms of the coinductive Omni-big-step judgment. *)

Definition coomnidiv (s:state) (t:trm) : Prop :=
  coomnibig s t Empty.


(* ########################################################### *)
(** ** Small-Step Divergence *)

(** The judgment [sdiv s t] characterize divergence in small-step:
    any execution prefix can be extended. *)

Definition sdiv (s:state) (t:trm) : Prop :=
  forall s2 t2, steps s t s2 t2 ->
  exists s3 t3, step s2 t2 s3 t3.

(** We prove the equivalence between [sdiv] and the specialization
    of [partial] to [Empty]. *)

Lemma sdiv_iff_omnibigps_Empty : forall s t,
  sdiv s t <-> partial s t Empty.
Proof using.
  iff M.
  { introv R. right. applys M R. }
  { introv R. lets C: M R. destruct C as [(v2&E&HQ)|C].
    { unfolds Empty. false. }
    { applys C. } }
Qed.


(* ########################################################### *)
(** ** CoInductive Divergence in Small-Step *)

(** [stepinf s t] is equivalent to [sdiv], but defined coinductively. *)

CoInductive stepinf : state -> trm -> Prop :=
  | stepinf_step : forall s1 t1,
      (exists s2 t2, step s1 t1 s2 t2) ->
      (forall s2 t2, step s1 t1 s2 t2 -> stepinf s2 t2) ->
      stepinf s1 t1.

(** We relate [stepinf] to [cobigsmall] with an empty postcondition. *)

Lemma stepinf_iff_cobigsmall_Empty : forall s t,
  stepinf s t <-> cobigsmall s t Empty.
Proof using.
  iff M.
  { gen s t. cofix IH. intros. inverts M as S M'.
    applys cobigsmall_step S. introv S'. applys IH. applys M' S'. }
  { gen s t. cofix IH. intros. inverts M as.
    { introv N. unfolds Empty. false. }
    { introv S M'. applys stepinf_step S.
      introv S'. applys IH. applys M' S'. } }
Qed.


(* ########################################################### *)
(** ** Equivalence Between Definitions of Divergence *)

(** We prove that [sdiv] matches [stepinf] and [coomnidiv]. *)

(** We establish the equivalence between the two small-step
    characterizations of divergence. *)

Lemma stepinf_eq_sdiv :
  stepinf = sdiv.
Proof using.
  extens. intros s t.
  rewrite sdiv_iff_omnibigps_Empty.
  rewrite stepinf_iff_cobigsmall_Empty.
  rewrite* cobigsmall_eq_partial.
Qed.

(** We establish the equivalence between the small-step characterisation
   and the coinductive Omni-big-step characterization of divergence. *)

Lemma sdiv_eq_coomnidiv :
  sdiv = coomnidiv.
Proof using.
  extens. intros s t. unfold coomnidiv.
  rewrite sdiv_iff_omnibigps_Empty.
  rewrite <- cobigsmall_eq_omnibigp.
  rewrite* cobigsmall_eq_partial.
Qed.


(* ########################################################### *)
(* ########################################################### *)
(* ########################################################### *)
(** * Hoare Triples *)

(* ########################################################### *)
(** ** Total Correctness *)

(** A total correctness Hoare triple, written [hoaret t H Q], asserts that,
    for any input state [s] satisfying the precondition [H], all executions of
    [t] in that state [s] terminate and produce an output satisfying the
    postcondition [Q]. *)

Definition hoaret (t:trm) (H:hprop) (Q:val->hprop) : Prop :=
  forall (s:state), H s -> omnibig s t Q.

(** Rules *)

Lemma hoaret_conseq : forall t H' Q' H Q,
  hoaret t H' Q' ->
  H ==> H' ->
  Q' ===> Q ->
  hoaret t H Q.
Proof using.
  unfolds hoaret. introv M MH MQ HF. applys* omnibig_conseq.
Qed.

Lemma hoaret_val : forall v H Q,
  H ==> Q v ->
  hoaret (trm_val v) H Q.
Proof using. introv M. intros h K. applys* omnibig_val. Qed.

Lemma hoaret_fix : forall f x t1 H Q,
  H ==> Q (val_fix f x t1) ->
  hoaret (trm_fix f x t1) H Q.
Proof using. introv M. intros h K. applys* omnibig_fix. Qed.

Lemma hoaret_app_fix : forall v1 v2 f x t1 H Q,
  v1 = val_fix f x t1 ->
  hoaret (subst x v2 (subst f v1 t1)) H Q ->
  hoaret (trm_app v1 v2) H Q.
Proof using. introv E M. intros h K. applys* omnibig_app_fix. Qed.

Lemma hoaret_let : forall x t1 t2 H Q Q1,
  hoaret t1 H Q1 ->
  (forall v1, hoaret (subst x v1 t2) (Q1 v1) Q) ->
  hoaret (trm_let x t1 t2) H Q.
Proof using. introv M1 M2. intros h K. applys* omnibig_let. Qed.

Lemma hoaret_if : forall (b:bool) t1 t2 H Q,
  hoaret (if b then t1 else t2) H Q ->
  hoaret (trm_if b t1 t2) H Q.
Proof using. introv M. intros h K. applys* omnibig_if. Qed.


(* ########################################################### *)
(** ** A Few Separation Logic Notation *)

(** These definitions are used to state the remaining Hoare
    Logic rules, in a way that paves the way for their Separation
    Logic counterpart. *)

Declare Scope hprop_scope.
Open Scope hprop_scope.

Definition hempty : hprop :=
  fun h => (h = Fmap.empty).

Notation "\[]" := (hempty)
  (at level 0) : hprop_scope.

Definition hsingle (p:loc) (v:val) : hprop :=
  fun h => (h = Fmap.single p v).

Notation "p '~~>' v" := (hsingle p v) (at level 32) : hprop_scope.

Definition hstar (H1 H2 : hprop) : hprop :=
  fun h => exists h1 h2, H1 h1
                              /\ H2 h2
                              /\ Fmap.disjoint h1 h2
                              /\ h = Fmap.union h1 h2.

Notation "H1 '\*' H2" := (hstar H1 H2)
  (at level 41, right associativity) : hprop_scope.

Definition hexists A (J:A->hprop) : hprop :=
  fun h => exists x, J x h.

Notation "'\exists' x1 .. xn , H" :=
  (hexists (fun x1 => .. (hexists (fun xn => H)) ..))
  (at level 39, x1 binder, H at level 50, right associativity,
   format "'[' '\exists' '/ '  x1  ..  xn , '/ '  H ']'") : hprop_scope.

Definition hpure (P:Prop) : hprop :=
  \exists (p:P), \[].

Notation "\[ P ]" := (hpure P)
  (at level 0, format "\[ P ]") : hprop_scope.

(** A few handy lemmas for introduction/elimination of SL operators. *)

Lemma hpure_intro : forall (P:Prop),
  P ->
  \[P] Fmap.empty.
Proof using. introv M. hnf. exists*. hnfs*. Qed.

Lemma hempty_inv : forall h,
  \[] h ->
  h = Fmap.empty.
Proof using. introv M. hnf in M. auto. Qed.

Lemma hexists_intro : forall A (x:A) (J:A->hprop) h,
  J x h ->
  (\exists x, J x) h.
Proof using. introv M. hnf. eauto. Qed.

Lemma hsingle_intro : forall p v,
  (p ~~> v) (Fmap.single p v).
Proof using. intros. hnf. auto. Qed.

Lemma hsingle_inv: forall p v h,
  (p ~~> v) h ->
  h = Fmap.single p v.
Proof using. introv M. hnf in M. auto. Qed.

Lemma hstar_intro : forall H1 H2 h1 h2,
  H1 h1 ->
  H2 h2 ->
  Fmap.disjoint h1 h2 ->
  (H1 \* H2) (h1 \u h2).
Proof using. intros. exists* h1 h2. Qed.

Lemma hstar_inv : forall H1 H2 h,
  (H1 \* H2) h ->
  exists h1 h2, H1 h1 /\ H2 h2 /\ Fmap.disjoint h1 h2 /\ h = h1 \u h2.
Proof using. introv M. hnf in M. eauto. Qed.

Lemma hstar_hpure_l : forall (P:Prop) H h,
  (\[P] \* H) h = (P /\ H h).
Proof using.
  intros. extens. iff M.
  { hnf in M. destruct M as (h1&h2&M1&M2&D&U).
    hnf in M1. destruct M1 as (M1&HP).
    lets: hempty_inv HP. subst.
    rewrite Fmap.union_empty_l. auto. }
  { destruct M as (HP&Hh). hnf.
    exists (Fmap.empty:state) h. splits.
    { applys* hpure_intro. }
    { auto. }
    { apply Fmap.disjoint_empty_l. }
    { rewrite Fmap.union_empty_l. auto. } }
Qed.

Lemma himpl_frame_l : forall H2 H1 H1',
  H1 ==> H1' ->
  (H1 \* H2) ==> (H1' \* H2).
Proof using. introv W (h1&h2&?). exists* h1 h2. Qed.

Lemma himpl_frame_r : forall H1 H2 H2',
  H2 ==> H2' ->
  (H1 \* H2) ==> (H1 \* H2').
Proof using. introv W (h1&h2&?). exists* h1 h2. Qed.

Lemma himpl_hfalse_l : forall H,
  \[False] ==> H.
Proof using. introv (N&HN). false. Qed.


(* ########################################################### *)
(** ** Proof of the Frame Rule *)

(** The key lemma is to argue that [omnibig] is stable by extension with a
    disjoint piece of heap. This proof is by induction. The most (and only)
    interesting step is that of allocation. In a derivation built using
    [omnibig_ref], we are given an assumption about all locations [p]
    fresh from [h1]. What we are trying to build is a fact mentioning all
    locations [p] fresh from [h1 \u h2]. We are thus restricting the set
    of [p] that can be considered for allocation. Hence the result holds. *)

Lemma omnibig_frame : forall h1 h2 t Q,
  omnibig h1 t Q ->
  Fmap.disjoint h1 h2 ->
  omnibig (h1 \u h2) t (fun r => Q r \* (= h2)).
Proof using.
  introv M HD. gen h2. induction M; intros;
    try solve [ hint hstar_intro; constructors* ].
  { rename M into M1, H into M2, IHM into IH1, H0 into IH2.
    specializes IH1 HD. applys omnibig_let IH1. introv HK.
    lets (h1'&h2'&K1'&K2'&KD&KU): hstar_inv HK. subst. applys* IH2. }
  { rename H into M. applys omnibig_ref. intros p Hp.
    rewrite Fmap.indom_union_eq in Hp. rew_logic in Hp. destruct Hp as [Hp1 Hp2].
    rewrite* Fmap.update_union_not_r. applys hstar_intro.
    { applys* M. } { auto. } { applys* Fmap.disjoint_update_not_r. } }
  { applys omnibig_get. { rewrite* Fmap.indom_union_eq. }
    { rewrite* Fmap.read_union_l. applys* hstar_intro. } }
  { applys omnibig_set. { rewrite* Fmap.indom_union_eq. }
    { rewrite* Fmap.update_union_l. applys hstar_intro.
      { auto. } { auto. } { applys* Fmap.disjoint_update_l. } } }
  { applys omnibig_free. { rewrite* Fmap.indom_union_eq. }
    { rewrite* Fmap.remove_disjoint_union_l. applys hstar_intro.
      { auto. } { auto. } { applys* Fmap.disjoint_remove_l. } } }
Qed.

(** From [omnibig_frame], we can derive the standard frame rule for [hoaret]. *)

Lemma hoaret_frame : forall t H Q H',
  hoaret t H Q ->
  hoaret t (H \* H') (fun r => Q r \* H').
Proof.
  introv M. intros h HF. lets (h1&h2&M1&M2&MD&MU): hstar_inv (rm HF).
  subst. specializes M M1. applys omnibig_conseq.
  { applys omnibig_frame M MD. }
  { intros r. applys himpl_frame_r. intros ? ->. applys M2. }
Qed.


(* ########################################################### *)
(** ** Other Separation Logic Rules *)

Lemma hoaret_hexists : forall t (A:Type) (J:A->hprop) Q,
  (forall x, hoaret t (J x) Q) ->
  hoaret t (hexists J) Q.
Proof using. introv M. intros h (x&Hh). applys M Hh. Qed.

Lemma hoaret_hpure : forall t (P:Prop) H Q,
  (P -> hoaret t H Q) ->
  hoaret t (\[P] \* H) Q.
Proof using.
  introv M. intros h (h1&h2&M1&M2&D&U). destruct M1 as (M1&HP).
  lets E: hempty_inv HP. subst. rewrite Fmap.union_empty_l. applys~ M.
Qed.

Lemma hoaret_add : forall H n1 n2,
  hoaret (val_add n1 n2)
    H
    (fun r => \[r = val_int (n1 + n2)] \* H).
Proof using.
  intros. intros h K. applys* omnibig_add. rewrite* hstar_hpure_l.
Qed.

Lemma hoaret_rand : forall n,
  n > 0 ->
  hoaret (val_rand n)
    \[]
    (fun r => \exists n1, \[0 <= n1 < n] \* \[r = val_int n1]).
Proof using.
  introv N. intros h K. applys* omnibig_rand.
  lets ->: hempty_inv K.
  intros n1 H1. applys* hexists_intro n1. rewrite* hstar_hpure_l.
  split*. applys* hpure_intro.
Qed.

Lemma hoaret_ref : forall H v,
  hoaret (val_ref v)
    H
    (fun r => (\exists p, \[r = val_loc p] \* p ~~> v) \* H).
Proof using.
  intros. intros s1 K. applys omnibig_ref. intros p D.
  unfolds update. applys hstar_intro K.
  { applys hexists_intro p. rewrite hstar_hpure_l.
    split*. applys hsingle_intro. }
  { applys* disjoint_single_of_not_indom. }
Qed.

Lemma hoaret_get : forall H v p,
  hoaret (val_get p)
    ((p ~~> v) \* H)
    (fun r => \[r = v] \* (p ~~> v) \* H).
Proof using.
  intros. intros s K.
  lets (h1&h2&P1&P2&D&->): hstar_inv (rm K).
  lets E1: hsingle_inv P1. lets I1: indom_single p v.
  applys omnibig_get.
  { applys* Fmap.indom_union_l. subst. applys indom_single. }
  { rewrite hstar_hpure_l. split.
    { subst h1. rewrite* Fmap.read_union_l. rewrite* Fmap.read_single. }
    { applys* hstar_intro. } }
Qed.

Lemma hoaret_set : forall H w p v,
  hoaret (val_set (val_loc p) v)
    ((p ~~> w) \* H)
    (fun r => \[r = val_unit] \* (p ~~> v) \* H).
Proof using.
  intros. intros s1 K.
  lets (h1&h2&P1&P2&D&->): hstar_inv (rm K).
  lets E1: hsingle_inv P1. lets I1: indom_single p v.
  applys omnibig_set.
  { applys* Fmap.indom_union_l. subst. applys indom_single. }
  { rewrite hstar_hpure_l. split*.
    { subst h1. rewrite* Fmap.update_union_l. rewrite* update_single.
      applys* hstar_intro.
      { applys* hsingle_intro. }
      { applys Fmap.disjoint_single_set D. }
      { applys indom_single. } } }
Qed.

Lemma hoaret_free : forall H p v,
  hoaret (val_free (val_loc p))
    ((p ~~> v) \* H)
    (fun r => \[r = val_unit] \* H).
Proof using.
  intros. intros s1 K.
  lets (h1&h2&P1&P2&D&->): hstar_inv (rm K).
  lets E1: hsingle_inv P1. lets I1: indom_single p v.
  applys omnibig_free.
  { applys* Fmap.indom_union_l. subst. applys indom_single. }
  { rewrite hstar_hpure_l. split*.
    { subst h1. rewrite* Fmap.remove_union_single_l.
      { intros Dl. applys* Fmap.disjoint_inv_not_indom_both D Dl. } } }
Qed.


(* ########################################################### *)
(** ** Partial Correctness Hoare and Separation Logic triples *)

(** The judgment [hoarep t H Q] captures partial correctness, in terms of
    [coomnibig]. *)

Definition hoarep (t:trm) (H:hprop) (Q:val->hprop) : Prop :=
  forall s, H s -> coomnibig s t Q.

(** Structural rules *)

Lemma hoarep_conseq : forall t H' Q' H Q,
  hoarep t H' Q' ->
  H ==> H' ->
  Q' ===> Q ->
  hoarep t H Q.
Proof using.
  unfolds hoarep. introv M MH MQ HF. applys* coomnibig_conseq.
Qed.

Lemma hoarep_hexists : forall t (A:Type) (J:A->hprop) Q,
  (forall x, hoarep t (J x) Q) ->
  hoarep t (hexists J) Q.
Proof using. introv M. intros h (x&Hh). applys M Hh. Qed.

Lemma hoarep_hpure : forall t (P:Prop) H Q,
  (P -> hoarep t H Q) ->
  hoarep t (\[P] \* H) Q.
Proof using.
  introv M. intros h (h1&h2&M1&M2&D&U). destruct M1 as (M1&HP).
  lets E: hempty_inv HP. subst. rewrite Fmap.union_empty_l. applys~ M.
Qed.

Lemma coomnibig_frame : forall h1 h2 t Q,
  coomnibig h1 t Q ->
  Fmap.disjoint h1 h2 ->
  coomnibig (h1 \u h2) t (fun r => Q r \* (= h2)).
Proof using.
  cofix IH. introv M HD. destruct M;
    try solve [ hint hstar_intro; constructors* ].
  { rename M into M1, H into M2. lets IH1: IH M1 HD.
    applys coomnibig_let IH1. introv HK.
    lets (h1'&h2'&K1'&K2'&KD&KU): hstar_inv HK. subst. applys* IH. }
  { rename H into M. applys coomnibig_ref. intros p Hp.
    rewrite Fmap.indom_union_eq in Hp. rew_logic in Hp. destruct Hp as [Hp1 Hp2].
    rewrite* Fmap.update_union_not_r. applys hstar_intro.
    { applys* M. } { auto. } { applys* Fmap.disjoint_update_not_r. } }
  { applys coomnibig_get. { rewrite* Fmap.indom_union_eq. }
    { rewrite* Fmap.read_union_l. applys* hstar_intro. } }
  { applys coomnibig_set. { rewrite* Fmap.indom_union_eq. }
    { rewrite* Fmap.update_union_l. applys hstar_intro.
      { auto. } { auto. } { applys* Fmap.disjoint_update_l. } } }
  { applys coomnibig_free. { rewrite* Fmap.indom_union_eq. }
    { rewrite* Fmap.remove_disjoint_union_l. applys hstar_intro.
      { auto. } { auto. } { applys* Fmap.disjoint_remove_l. } } }
Qed.

(** From [omnibig_frame], we can derive the standard frame rule for [hoaret]. *)

Lemma hoarep_frame : forall t H Q H',
  hoarep t H Q ->
  hoarep t (H \* H') (fun r => Q r \* H').
Proof.
  introv M. intros h HF. lets (h1&h2&M1&M2&MD&MU): hstar_inv (rm HF).
  subst. specializes M M1. applys coomnibig_conseq.
  { applys coomnibig_frame M MD. }
  { intros r. applys himpl_frame_r. intros ? ->. applys M2. }
Qed.

(** Rules for term constructs *)

Lemma hoarep_val : forall v H Q,
  H ==> Q v ->
  hoarep (trm_val v) H Q.
Proof using. introv M. intros h K. applys* coomnibig_val. Qed.

Lemma hoarep_fix : forall f x t1 H Q,
  H ==> Q (val_fix f x t1) ->
  hoarep (trm_fix f x t1) H Q.
Proof using. introv M. intros h K. applys* coomnibig_fix. Qed.

Lemma hoarep_app_fix : forall v1 v2 f x t1 H Q,
  v1 = val_fix f x t1 ->
  hoarep (subst x v2 (subst f v1 t1)) H Q ->
  hoarep (trm_app v1 v2) H Q.
Proof using. introv E M. intros h K. applys* coomnibig_app_fix. Qed.

Lemma hoarep_let : forall x t1 t2 H Q Q1,
  hoarep t1 H Q1 ->
  (forall v1, hoarep (subst x v1 t2) (Q1 v1) Q) ->
  hoarep (trm_let x t1 t2) H Q.
Proof using. introv M1 M2. intros h K. applys* coomnibig_let. Qed.

Lemma hoarep_if : forall (b:bool) t1 t2 H Q,
  hoarep (if b then t1 else t2) H Q ->
  hoarep (trm_if b t1 t2) H Q.
Proof using. introv M. intros h K. applys* coomnibig_if. Qed.

(** Rules for primitive operations *)

Lemma hoarep_add : forall H n1 n2,
  hoarep (val_add n1 n2)
    H
    (fun r => \[r = val_int (n1 + n2)] \* H).
Proof using.
  intros. intros h K. applys* coomnibig_add. rewrite* hstar_hpure_l.
Qed.

Lemma hoarep_rand : forall n,
  n > 0 ->
  hoarep (val_rand n)
    \[]
    (fun r => \exists n1, \[0 <= n1 < n] \* \[r = val_int n1]).
Proof using.
  introv N. intros h K. applys* coomnibig_rand.
  lets ->: hempty_inv K.
  intros n1 H1. applys* hexists_intro n1. rewrite* hstar_hpure_l.
  split*. applys* hpure_intro.
Qed.

Lemma hoarep_ref : forall H v,
  hoarep (val_ref v)
    H
    (fun r => (\exists p, \[r = val_loc p] \* p ~~> v) \* H).
Proof using.
  intros. intros s1 K. applys coomnibig_ref. intros p D.
  unfolds update. applys hstar_intro K.
  { applys hexists_intro p. rewrite hstar_hpure_l.
    split*. applys hsingle_intro. }
  { applys* disjoint_single_of_not_indom. }
Qed.

Lemma hoarep_get : forall H v p,
  hoarep (val_get p)
    ((p ~~> v) \* H)
    (fun r => \[r = v] \* (p ~~> v) \* H).
Proof using.
  intros. intros s K.
  lets (h1&h2&P1&P2&D&->): hstar_inv (rm K).
  lets E1: hsingle_inv P1. lets I1: indom_single p v.
  applys coomnibig_get.
  { applys* Fmap.indom_union_l. subst. applys indom_single. }
  { rewrite hstar_hpure_l. split.
    { subst h1. rewrite* Fmap.read_union_l. rewrite* Fmap.read_single. }
    { applys* hstar_intro. } }
Qed.

Lemma hoarep_set : forall H w p v,
  hoarep (val_set (val_loc p) v)
    ((p ~~> w) \* H)
    (fun r => \[r = val_unit] \* (p ~~> v) \* H).
Proof using.
  intros. intros s1 K.
  lets (h1&h2&P1&P2&D&->): hstar_inv (rm K).
  lets E1: hsingle_inv P1. lets I1: indom_single p v.
  applys coomnibig_set.
  { applys* Fmap.indom_union_l. subst. applys indom_single. }
  { rewrite hstar_hpure_l. split*.
    { subst h1. rewrite Fmap.update_union_l.
      2:{ applys indom_single. }
      rewrite update_single. applys hstar_intro.
      { applys hsingle_intro. }
      { applys P2. }
      { applys Fmap.disjoint_single_set D. } } }
Qed.

Lemma hoarep_free : forall H p v,
  hoarep (val_free (val_loc p))
    ((p ~~> v) \* H)
    (fun r => \[r = val_unit] \* H).
Proof using.
  intros. intros s1 K.
  lets (h1&h2&P1&P2&D&->): hstar_inv (rm K).
  lets E1: hsingle_inv P1. lets I1: indom_single p v.
  applys coomnibig_free.
  { applys* Fmap.indom_union_l. subst. applys indom_single. }
  { rewrite hstar_hpure_l. split*.
    { subst h1. rewrite* Fmap.remove_union_single_l.
      { intros Dl. applys* Fmap.disjoint_inv_not_indom_both D Dl. } } }
Qed.


(* ########################################################### *)
(* ########################################################### *)
(* ########################################################### *)
(** * Weakest-Preconditions *)

(* ########################################################### *)
(** ** Total Correctness Weakest-Preconditions *)

(** The judgment [wp t Q], defined below, simply reorder the arguments
    of [omnibig] so as to produce this weakest-precondition operator. Note that
    this is a Hoare Logic style predicate, which talks about the full state. *)

Definition wp (t:trm) (Q:val->hprop) : hprop :=
  fun s => omnibig s t Q.

(** Structural rules *)

Lemma wp_conseq : forall t Q1 Q2,
  Q1 ===> Q2 ->
  (wp t Q1) ==> (wp t Q2).
Proof using.
  introv W. unfold wp. intros h K. applys omnibig_conseq K W.
Qed.

Lemma wp_frame : forall t H Q,
  (wp t Q) \* H ==> wp t (fun r => Q r \* H).
Proof using.
  intros. unfold wp. intros h HF.
  lets (h1&h2&M1&M2&MD&MU): hstar_inv (rm HF).
  subst. applys omnibig_conseq.
  { applys omnibig_frame M1 MD. }
  { intros r. applys himpl_frame_r. intros ? ->. applys M2. }
Qed.

(** Rules for term constructs *)

Lemma wp_val : forall v Q,
  Q v ==> wp (trm_val v) Q.
Proof using.
  unfold wp. intros. intros h K. applys* omnibig_val.
Qed.

Lemma wp_fix : forall f x t Q,
  Q (val_fix f x t) ==> wp (trm_fix f x t) Q.
Proof using.
  unfold wp. intros. intros h K. applys* omnibig_fix.
Qed.

Lemma wp_app_fix : forall f x v1 v2 t1 Q,
  v1 = val_fix f x t1 ->
  wp (subst x v2 (subst f v1 t1)) Q ==> wp (trm_app v1 v2) Q.
Proof using.
  unfold wp. intros. intros h K. applys* omnibig_app_fix.
Qed.

Lemma wp_let : forall x t1 t2 Q,
  wp t1 (fun v => wp (subst x v t2) Q) ==> wp (trm_let x t1 t2) Q.
Proof using.
  unfold wp. intros. intros h K. applys* omnibig_let.
Qed.

Lemma wp_if : forall b t1 t2 Q,
  wp (if b then t1 else t2) Q ==> wp (trm_if b t1 t2) Q.
Proof using.
  unfold wp. intros. intros h K. applys* omnibig_if.
Qed.

(* ########################################################### *)
(** ** Bonus: Ramified Frame Rule *)

(** Universal quantifier, written [\forall x, H] *)

Definition hforall (A : Type) (J : A -> hprop) : hprop :=
  fun h => forall x, J x h.

Notation "'\forall' x1 .. xn , H" :=
  (hforall (fun x1 => .. (hforall (fun xn => H)) ..))
  (at level 39, x1 binder, H at level 50, right associativity,
   format "'[' '\forall' '/ '  x1  ..  xn , '/ '  H ']'") : heap_scope.

(** Magic wand, written [H1 \-* H2] *)

Definition hwand (H1 H2 : hprop) : hprop :=
  hexists (fun (H:hprop) => (hpure (H1 \* H ==> H2) \* H)).

Notation "H1 \-* H2" := (hwand H1 H2)
  (at level 43, right associativity) : hprop_scope.

(** Magic wand for postconditions, written [Q1 \--* Q2] *)

Definition qwand A (Q1 Q2:A->hprop) : hprop :=
  hforall (fun x => hwand (Q1 x) (Q2 x)).

Notation "Q1 \--* Q2" := (qwand Q1 Q2)
  (at level 43) : hprop_scope.

(** Ramified frame rule *)

Lemma wp_ramified_frame : forall t Q1 Q2,
  (wp t Q1) \* (Q1 \--* Q2) ==> (wp t Q2).
Proof using.
  intros. applys himpl_trans.
  { applys wp_frame. }
  { applys wp_conseq. intros r.
    unfold qwand, hwand, hforall. intros h (h1&h2&M1&M2&D&U).
    specializes M2 r. destruct M2 as (H'&MH').
    rewrite hstar_hpure_l in MH'. destruct MH' as (MI&M2).
    applys MI. exists* h1 h2. }
Qed.


(* ########################################################### *)
(* ########################################################### *)
(* ########################################################### *)
(** * Beyond A-normal form: Adding the Bind Rule  *)

Module Bind.

(* ########################################################### *)
(** ** Omni-Big-Step Judgment with Evaluation under Context *)

(** Definition of evaluation contexts *)

Inductive ctx : Type :=
  | ctx_hole : ctx
  | ctx_let_1 : var -> ctx -> trm -> ctx
  | ctx_app_1 : ctx -> trm -> ctx
  | ctx_app_2 : val -> ctx -> ctx
  | ctx_if_1 : ctx -> trm -> trm -> ctx.

Fixpoint ctx_app (E:ctx) (u:trm) : trm :=
  match E with
  | ctx_hole => u
  | ctx_let_1 x E1 t2 => trm_let x (ctx_app E1 u) t2
  | ctx_app_1 E1 t2 => trm_app (ctx_app E1 u) t2
  | ctx_app_2 v1 E2 => trm_app (trm_val v1) (ctx_app E2 u)
  | ctx_if_1 E0 t1 t2 => trm_if (ctx_app E0 u) t1 t2
  end.

Definition isvalue (t:trm) : Prop :=
  match t with
  | trm_val v => True
  | _ => False
  end.

(** Judgment [omnibig s t Q] *)

Inductive omnibig : state -> trm -> (val->state->Prop) -> Prop :=
  | omnibig_bind : forall Q1 s t1 E Q,
      ~ isvalue t1 ->
      omnibig s t1 Q1 ->
      (forall v1 s2, Q1 v1 s2 -> omnibig s2 (ctx_app E (trm_val v1)) Q) ->
      omnibig s (ctx_app E t1) Q
  | omnibig_val : forall s v Q,
      Q v s ->
      omnibig s (trm_val v) Q
  | omnibig_fix : forall s f x t1 Q,
      Q (val_fix f x t1) s ->
      omnibig s (trm_fix f x t1) Q
  | omnibig_app_fix : forall s v1 v2 f x t1 Q,
      v1 = val_fix f x t1 ->
      omnibig s (subst x v2 (subst f v1 t1)) Q ->
      omnibig s (trm_app v1 v2) Q
  | omnibig_let : forall s x v1 t2 Q,
      omnibig s (subst x v1 t2) Q ->
      omnibig s (trm_let x v1 t2) Q
  | omnibig_if : forall s b t1 t2 Q,
      omnibig s (if b then t1 else t2) Q ->
      omnibig s (trm_if (val_bool b) t1 t2) Q
  | omnibig_add : forall s n1 n2 Q,
      Q (val_int (n1 + n2)) s ->
      omnibig s (val_add (val_int n1) (val_int n2)) Q
  | omnibig_rand : forall s n Q,
      n > 0 ->
      (forall n1, 0 <= n1 < n -> Q n1 s) ->
      omnibig s (val_rand (val_int n)) Q
  | omnibig_ref : forall s v Q,
      (forall p, ~ Fmap.indom s p ->
         Q (val_loc p) (Fmap.update s p v)) ->
      omnibig s (val_ref v) Q
  | omnibig_get : forall s p Q,
      Fmap.indom s p ->
      Q (Fmap.read s p) s ->
      omnibig s (val_get (val_loc p)) Q
  | omnibig_set : forall s p v Q,
      Fmap.indom s p ->
      Q val_unit (Fmap.update s p v) ->
      omnibig s (val_set (val_loc p) v) Q
  | omnibig_free : forall s p Q,
      Fmap.indom s p ->
      Q val_unit (Fmap.remove s p) ->
      omnibig s (val_free (val_loc p)) Q.

Lemma omnibig_conseq : forall s t Q1 Q2,
  omnibig s t Q1 ->
  Q1 ===> Q2 ->
  omnibig s t Q2.
Proof using.
  introv M W.
  asserts W': (forall v h, Q1 v h -> Q2 v h). { auto. } clear W.
  induction M; try solve [ constructors* ].
Qed.

Lemma omnibig_frame : forall h1 h2 t Q,
  omnibig h1 t Q ->
  Fmap.disjoint h1 h2 ->
  omnibig (h1 \u h2) t (fun r => Q r \* (= h2)).
Proof using.
  introv M HD. gen h2. induction M; intros;
    try solve [ hint hstar_intro; constructors* ].
  { rename M into M1, H into NV1, H0 into M2, IHM into IH1, H1 into IH2.
    specializes IH1 HD. applys omnibig_bind NV1 IH1. introv HK.
    lets (h1'&h2'&K1'&K2'&KD&KU): hstar_inv HK. subst. applys* IH2. }
  { rename H into M. applys omnibig_ref. intros p Hp.
    rewrite Fmap.indom_union_eq in Hp. rew_logic in Hp. destruct Hp as [Hp1 Hp2].
    rewrite* Fmap.update_union_not_r. applys hstar_intro.
    { applys* M. } { auto. } { applys* Fmap.disjoint_update_not_r. } }
  { applys omnibig_get. { rewrite* Fmap.indom_union_eq. }
    { rewrite* Fmap.read_union_l. applys* hstar_intro. } }
  { applys omnibig_set. { rewrite* Fmap.indom_union_eq. }
    { rewrite* Fmap.update_union_l. applys hstar_intro.
      { auto. } { auto. } { applys* Fmap.disjoint_update_l. } } }
  { applys omnibig_free. { rewrite* Fmap.indom_union_eq. }
    { rewrite* Fmap.remove_disjoint_union_l. applys hstar_intro.
      { auto. } { auto. } { applys* Fmap.disjoint_remove_l. } } }
Qed.


(* ########################################################### *)
(** ** Weakest-Preconditions with a Bind rule *)

(** The judgment [wp t Q], defined below, simply reorder the arguments
    of [omnibig] so as to produce this weakest-precondition operator. Note that
    this is a Hoare Logic style predicate, which talks about the full state. *)

Definition wp (t:trm) (Q:val->hprop) : hprop :=
  fun s => omnibig s t Q.

(** Structural rules *)

Lemma wp_conseq : forall t Q1 Q2,
  Q1 ===> Q2 ->
  (wp t Q1) ==> (wp t Q2).
Proof using.
  introv W. unfold wp. intros h K. applys omnibig_conseq K W.
Qed.

Lemma wp_frame : forall t H Q,
  (wp t Q) \* H ==> wp t (fun r => Q r \* H).
Proof using.
  intros. unfold wp. intros h HF.
  lets (h1&h2&M1&M2&MD&MU): hstar_inv (rm HF).
  subst. applys omnibig_conseq.
  { applys omnibig_frame M1 MD. }
  { intros r. applys himpl_frame_r. intros ? ->. applys M2. }
Qed.

(** Rules for term constructs *)

Lemma wp_val : forall v Q,
  Q v ==> wp (trm_val v) Q.
Proof using.
  unfold wp. intros. intros h K. applys* omnibig_val.
Qed.

Lemma wp_fix : forall f x t Q,
  Q (val_fix f x t) ==> wp (trm_fix f x t) Q.
Proof using.
  unfold wp. intros. intros h K. applys* omnibig_fix.
Qed.

Lemma wp_app_fix : forall f x v1 v2 t1 Q,
  v1 = val_fix f x t1 ->
  wp (subst x v2 (subst f v1 t1)) Q ==> wp (trm_app v1 v2) Q.
Proof using.
  unfold wp. intros. intros h K. applys* omnibig_app_fix.
Qed.

Lemma wp_let : forall x v1 t2 Q,
  wp (subst x v1 t2) Q ==> wp (trm_let x v1 t2) Q.
Proof using.
  unfold wp. intros. intros h K. applys* omnibig_let.
Qed.

Lemma wp_if : forall b t1 t2 Q,
  wp (if b then t1 else t2) Q ==> wp (trm_if b t1 t2) Q.
Proof using.
  unfold wp. intros. intros h K. applys* omnibig_if.
Qed.

(* Useful rule: inversion rule for wp for values *)

Lemma wp_val_inv : forall Q v,
  wp v Q ==> Q v.
Proof using.
  introv M. inverts M as.
  { intros Q1 t1 E N1 _ _ EQ. false. induction E; tryfalse; simpls.
    { subst. false N1. unfolds*. } }
  { auto. }
Qed. 

(* General bind rule: handles uniformly values and non-values *)

Lemma wp_bind : forall E t Q,
  wp t (fun v => wp (ctx_app E v) Q) ==> wp (ctx_app E t) Q.
Proof using.
  intros. tests C: (~ isvalue t).
  { unfold wp. intros. intros h K. applys* omnibig_bind. }
  { destruct t; tryfalse. applys himpl_trans.
    { applys wp_val_inv. } { applys himpl_refl. } }
Qed.

(* Derived rule: bind+let *)

Lemma wp_let_trm : forall x t1 t2 Q,
  wp t1 (fun v => wp (subst x v t2) Q) ==> wp (trm_let x t1 t2) Q.
Proof using.
  intros. applys himpl_trans; [ | applys* (wp_bind (ctx_let_1 x ctx_hole t2)) ].
  applys wp_conseq. intros v. simpl. applys wp_let.
Qed.

(* Derived rule: bind+if *)

Definition post_bool (Q:bool->hprop) : val->hprop :=
  fun v =>
    match v with
    | val_bool b => Q b
    | _ => \[False]
    end. 

Lemma wp_if_trm : forall t0 t1 t2 Q,
  wp t0 (post_bool (fun b => wp (if b then t1 else t2) Q)) ==> wp (trm_if t0 t1 t2) Q.
Proof using.
  intros. applys himpl_trans; [ | applys* (wp_bind (ctx_if_1 ctx_hole t1 t2)) ].
  applys wp_conseq. intros v.  
  unfold post_bool. destruct v; try apply himpl_hfalse_l. { applys wp_if. }
Qed.

(* Derived rule: bind+bind+app *)

Definition post_fix (Q:var->var->trm->hprop) : val->hprop :=
  fun v =>
    match v with
    | val_fix f x t1 => Q f x t1
    | _ => \[False]
    end. 

Lemma wp_app_trm : forall t1 t2 Q,
      wp t1 (post_fix (fun f x t1 => 
        wp t2 (fun v2 => 
          wp (subst x v2 (subst f (val_fix f x t1) t1)) Q))) 
  ==> wp (trm_app t1 t2) Q.
Proof using.
  intros. applys himpl_trans; [ | applys* (wp_bind (ctx_app_1 ctx_hole t2)) ].
  applys wp_conseq. intros v1. simpl.
  unfold post_fix. destruct v1; try apply himpl_hfalse_l.
  clear t1. rename v into f, v0 into x, t into t1.
  applys himpl_trans; [ | applys (wp_bind (ctx_app_2 (val_fix f x t1) ctx_hole)) ].
  applys wp_conseq. intros v2. simpl. applys* wp_app_fix.
Qed.

(* Other rules can be handled similarly *)

(* Deriving the bind rule for triples *)

Definition hoaret (t:trm) (H:hprop) (Q:val->hprop) : Prop :=
  forall (s:state), H s -> omnibig s t Q.

Lemma omnibig_val_inv : forall s Q v,
  omnibig s v Q ->
  Q v s.
Proof using.
  introv M. inverts M. 
  { false. induction E; tryfalse; simpls. { subst. false*. } }
  { auto. }
Qed. 

Lemma hoaret_bind : forall E t H Q Q1,
  hoaret t H Q1 ->
  (forall v, hoaret (ctx_app E v) (Q1 v) Q) ->
  hoaret (ctx_app E t) H Q.
Proof using.
  introv M1 M2. intros h K. tests C: (~ isvalue t).
  { applys* omnibig_bind. }
  { destruct t; tryfalse. unfolds hoaret. specializes M1 K.
    lets R1: omnibig_val_inv M1. applys M2 R1. }
Qed.

End Bind.


(* ########################################################### *)
(* ########################################################### *)
(* ########################################################### *)
(** * Compiler Correctness For Terminating Programs *)

Section CompilerCorrectnessTerminating.

(** Consider a compilation function. For simplicity, we assume
    here that the corresponding compilation phase does not affect
    the memory state during program executions (this is in particular
    the case for compilation of purely functional programs). *)

Parameter compile : trm -> trm.

(** Assume that it satisfies Omni-forward-preservation. *)

Parameter omni_forward_preservation_for_compile : forall s t Q,
  omnibig s t Q ->
  omnibig s (compile t) Q.

(** We show that Omni-forward-preservation entails correctness
    for terminating programs. *)

Lemma compile_correct : forall s t,
  terminates s t ->
     terminates s (compile t)
  /\ (forall s' v, big s (compile t) s' v -> big s t s' v).
Proof using.
  introv HT.
  forwards R: omni_forward_preservation_for_compile s t (strongestpost s t).
  { applys omnibig_strongestpost_of_terminates HT. }
  split.
  { applys terminates_of_omnibig R. }
  { introv M. applys omnibig_and_big_inv R M. }
Qed.

End CompilerCorrectnessTerminating.
