Set Implicit Arguments.
From DEV Require Export LibLN LibNat.

(** The most interesting lemmas are
    [omnismall_soundness_induction] and
    [coomnibig_soundness_coinduction].
    Each of them should be compared to the pair of
    [preservation_result] and [progress_result], which
    correspond to the standard small-step type soundness proof. *)


(* ########################################################### *)
(* ########################################################### *)
(* ########################################################### *)
(** * Language *)

(** We consider a language in A-normal form. In this file,
    instead of introducing let-bindings, we view [let x = t1 in t2]
    as syntactic sugar for [app (abs x => t2) t1], and we thus only
    have to consider right-branches of applications as reduction context.
    This design choice was motivated for minimizing the changes w.r.t.
    the original formalization of type soundness proof from the
    locally nameless sample developments. *)

(** Grammar of types. *)

Parameter atm : Set.

Inductive typ : Set :=
  | typ_var   : var -> typ
  | typ_arrow : typ -> typ -> typ
  | typ_ref   : typ -> typ
  | typ_int   : typ
  | typ_unit  : typ.

(** Grammar of pre-terms. *)

Definition loc := var.

(** For simplicity, let's assume nonnegative integers *)

Definition int := nat.

Inductive trm : Set :=
  | trm_bvar : nat -> trm
  | trm_fvar : var -> trm
  | trm_abs  : trm -> trm
  | trm_app  : trm -> trm -> trm
  | trm_unit : trm
  | trm_int : int -> trm
  | trm_loc : loc -> trm
  | trm_ref : trm -> trm
  | trm_get : trm -> trm
  | trm_set : trm -> trm -> trm
  | trm_rand : trm -> trm.

(** Opening up abstractions *)

Fixpoint open_rec (k : nat) (u : trm) (t : trm) {struct t} : trm :=
  match t with
  | trm_bvar i    => If k = i then u else (trm_bvar i)
  | trm_fvar x    => trm_fvar x
  | trm_abs t1    => trm_abs (open_rec (S k) u t1)
  | trm_app t1 t2 => trm_app (open_rec k u t1) (open_rec k u t2)
  | trm_unit      => trm_unit
  | trm_int n     => trm_int n
  | trm_loc l     => trm_loc l
  | trm_ref t1    => trm_ref (open_rec k u t1)
  | trm_get t1    => trm_get (open_rec k u t1)
  | trm_set t1 t2 => trm_set (open_rec k u t1) (open_rec k u t2)
  | trm_rand t1   => trm_rand (open_rec k u t1)
  end.

Definition open t u := open_rec 0 u t.

Notation "{ k ~> u } t" := (open_rec k u t) (at level 67).
Notation "t ^^ u" := (open t u) (at level 67).
Notation "t ^ x" := (open t (trm_fvar x)).

(** Terms are locally-closed pre-terms *)

Inductive term : trm -> Prop :=
  | term_var : forall x,
      term (trm_fvar x)
  | term_abs : forall L t1,
      (forall x, x \notin L -> term (t1 ^ x)) ->
      term (trm_abs t1)
  | term_app : forall t1 t2,
      term t1 -> term t2 -> term (trm_app t1 t2)
  | term_trm_unit :
      term trm_unit
  | term_int : forall n,
      term (trm_int n)
  | term_loc : forall l,
      term (trm_loc l)
  | term_ref : forall t1,
      term t1 ->
      term (trm_ref t1)
  | term_get : forall t1,
      term t1 ->
      term (trm_get t1)
  | term_set : forall t1 t2,
      term t1 ->
      term t2 ->
      term (trm_set t1 t2)
  | term_rand : forall t1,
      term t1 ->
      term (trm_rand t1).

(** Store maps (sto) locations to values *)

Definition sto := LibEnv.env trm.

(** Store typing (phi) maps locations to type. *)

Definition phi := LibEnv.env typ.

(** Well-formed stores *)
Inductive sto_ok : sto -> Prop :=
  | sto_ok_empty : sto_ok empty
  | sto_ok_push : forall mu l t,
      sto_ok mu -> term t ->
      sto_ok (mu & l ~ t).

(** Environment is an associative list mapping variables to types. *)

Definition env := LibEnv.env typ.

(** Definition of values *)

Inductive value : trm -> Prop :=
  | value_abs : forall t1,
      term (trm_abs t1) ->
      value (trm_abs t1)
  | value_trm_unit :
      value trm_unit
  | value_loc : forall l,
      value (trm_loc l)
  | value_int : forall n,
      value (trm_int n).

(** Value or variable predicate---for A-normal form. *)

Definition val_or_var (t:trm) : Prop :=
     value t
  \/ (exists x, t = trm_fvar x).

(** Typing relation *)

Reserved Notation "E ! Y |= t ~: T" (at level 69).

Inductive typing : env -> phi -> trm -> typ -> Prop :=
  | typing_var : forall E Y x T,
      ok E ->
      binds x T E ->
      E ! Y |= (trm_fvar x) ~: T
  | typing_abs : forall L E Y U T t1,
      (forall x, x \notin L -> (E & x ~ U) ! Y  |= t1 ^ x ~: T) ->
      E ! Y |= (trm_abs t1) ~: (typ_arrow U T)
  | typing_app : forall T1 T2 E Y t1 t2,
      val_or_var t1 ->
      E ! Y |= t1 ~: (typ_arrow T1 T2) -> E ! Y  |= t2 ~: T1 ->
      E ! Y  |= (trm_app t1 t2) ~: T2
  | typing_unit : forall E Y,
      ok E ->
      E ! Y |= trm_unit ~: typ_unit
  | typing_int : forall E Y n,
      ok E ->
      E ! Y |= (trm_int n) ~: typ_int
  | typing_loc : forall E Y l T,
      ok E ->
      binds l T Y ->
      E ! Y |= (trm_loc l) ~: (typ_ref T)
  | typing_ref : forall E Y t1 T,
      val_or_var t1 ->
      E ! Y |= t1 ~: T ->
      E ! Y |= (trm_ref t1) ~: (typ_ref T)
  | typing_get : forall E Y t1 T,
      val_or_var t1 ->
      E ! Y |= t1 ~: (typ_ref T) ->
      E ! Y |= (trm_get t1) ~: T
  | typing_set : forall E Y t1 t2 T,
      val_or_var t1 ->
      val_or_var t2 ->
      E ! Y |= t1 ~: (typ_ref T) ->
      E ! Y |= t2 ~: T ->
      E ! Y |= (trm_set t1 t2) ~: typ_unit
  | typing_rand : forall E Y t1,
      val_or_var t1 ->
      E ! Y |= t1 ~: typ_int ->
      E ! Y |= (trm_rand t1) ~: typ_int

where "E ! Y |= t ~: T" := (typing E Y t T).

(** Typing judgment for stores *)

Definition sto_typing Y mu :=
     sto_ok mu
  /\ (forall l, l # mu -> l # Y)
  /\ (forall l T, binds l T Y ->
        exists t, binds l t mu
               /\ empty ! Y |= t ~: T).

Notation "Y |== mu" := (sto_typing Y mu) (at level 68).

(** A configuration is a pair of a term and a store *)

Definition conf := (trm * sto)%type.

(** Small-step reduction, in call-by-value *)

Inductive red : conf -> conf -> Prop :=

  | red_beta : forall mu t1 t2,
      sto_ok mu ->
      term (trm_abs t1) ->
      value t2 ->
      red (trm_app (trm_abs t1) t2, mu) (t1 ^^ t2, mu)
  | red_ref : forall mu t1 l,
      sto_ok mu ->
      value t1 ->
      l # mu ->
      red (trm_ref t1, mu) (trm_loc l, (mu & l ~ t1))
  | red_get : forall mu l t,
      sto_ok mu ->
      binds l t mu ->
      red (trm_get (trm_loc l), mu) (t, mu)
  | red_set : forall mu l t2,
      sto_ok mu ->
      value t2 ->
      red (trm_set (trm_loc l) t2, mu) (trm_unit, mu & l ~ t2)
  | red_rand : forall mu n m,
      sto_ok mu ->
      (0 <= m < max 1 n) ->
      red (trm_rand (trm_int n), mu) ((trm_int m), mu)
  | red_app_2 : forall mu mu' t1 t2 t2',
      value t1 ->
      red (t2, mu) (t2', mu') ->
      red (trm_app t1 t2, mu) (trm_app t1 t2', mu').

Notation "c --> c'" := (red c c') (at level 68).


(* ########################################################### *)
(* ########################################################### *)
(* ########################################################### *)
(** * Infrastructure *)


(* ********************************************************************** *)
(** ** Additional Definitions used in the Proofs *)

(** Computing free variables of a term. *)

Fixpoint fv (t : trm) {struct t} : vars :=
  match t with
  | trm_bvar i    => \{}
  | trm_fvar x    => \{x}
  | trm_abs t1    => (fv t1)
  | trm_app t1 t2 => (fv t1) \u (fv t2)
  | trm_unit      => \{}
  | trm_int n     => \{}
  | trm_loc l     => \{}
  | trm_ref t1    => (fv t1)
  | trm_get t1    => (fv t1)
  | trm_set t1 t2 => (fv t1) \u (fv t2)
  | trm_rand t1   => (fv t1)
  end.

(** Substitution from variables to values *)

Fixpoint subst (z : var) (u : trm) (t : trm) {struct t} : trm :=
  match t with
  | trm_bvar i    => trm_bvar i
  | trm_fvar x    => If x = z then u else (trm_fvar x)
  | trm_abs t1    => trm_abs (subst z u t1)
  | trm_app t1 t2 => trm_app (subst z u t1) (subst z u t2)
  | trm_unit      => trm_unit
  | trm_int n     => trm_int n
  | trm_loc l     => trm_loc l
  | trm_ref t1    => trm_ref (subst z u t1)
  | trm_get t1    => trm_get (subst z u t1)
  | trm_set t1 t2 => trm_set (subst z u t1) (subst z u t2)
  | trm_rand t1   => trm_rand (subst z u t1)
  end.

Notation "[ z ~> u ] t" := (subst z u t) (at level 68).

(** Definition of the body of an abstraction *)

Definition body t :=
  exists L, forall x, x \notin L -> term (t ^ x).


(* ********************************************************************** *)
(** ** Instanciation of Locally Nameless tactics *)

Ltac gather_vars :=
  let A := gather_vars_with (fun x : vars => x) in
  let B := gather_vars_with (fun x : var => \{x}) in
  let C := gather_vars_with (fun x : env => dom x) in
  let D := gather_vars_with (fun x : trm => fv x) in
  constr:(A \u B \u C \u D).

Ltac pick_fresh Y :=
  let L := gather_vars in (pick_fresh_gen L Y).

Tactic Notation "apply_fresh" constr(T) "as" ident(x) :=
  apply_fresh_base T gather_vars x.

Tactic Notation "apply_fresh" "*" constr(T) "as" ident(x) :=
  apply_fresh T as x; autos*.

Hint Constructors term sto_ok value red.


(* ********************************************************************** *)
(** ** Properties of substitution *)

(** Substitution on indices is identity on well-formed terms. *)

Lemma open_rec_term_core :forall t j v i u, i <> j ->
  {j ~> v}t = {i ~> u}({j ~> v}t) -> t = {i ~> u}t.
Proof.
  induction t; introv Neq Equ;
  simpl in *; inversion* Equ; f_equal*.
  case_nat*. case_nat*.
Qed.

Lemma open_rec_term : forall t u,
  term t -> forall k, t = {k ~> u}t.
Proof.
  induction 1; intros; simpl; f_equal*. unfolds open.
  pick_fresh x. apply* (@open_rec_term_core t1 0 (trm_fvar x)).
Qed.

(** Substitution for a fresh name is identity. *)

Lemma subst_fresh : forall x t u,
  x \notin fv t ->  [x ~> u] t = t.
Proof.
  intros. induction t; simpls; f_equal*. case_var*.
Qed.

(** Substitution distributes on the open operation. *)

Lemma subst_open : forall x u t1 t2, term u ->
  [x ~> u] (t1 ^^ t2) = ([x ~> u]t1) ^^ ([x ~> u]t2).
Proof.
  intros. unfold open. generalize 0.
  induction t1; intros; simpl; f_equal*.
  case_nat*. case_var*. apply* open_rec_term.
Qed.

(** Substitution and open_var for distinct names commute. *)

Lemma subst_open_var : forall x y u t, y <> x -> term u ->
  ([x ~> u]t) ^ y = [x ~> u] (t ^ y).
Proof.
  introv Neq Wu. rewrite* subst_open.
  simpl. case_var*.
Qed.

(** Opening up an abstraction of body t with a term u is the same as opening
  up the abstraction with a fresh name x and then substituting u for x. *)

Lemma subst_intro : forall x t u,
  x \notin (fv t) -> term u ->
  t ^^ u = [x ~> u](t ^ x).
Proof.
  introv Fr Wu. rewrite* subst_open.
  rewrite* subst_fresh. simpl. case_var*.
Qed.


(* ********************************************************************** *)
(** ** Terms are stable through substitutions *)

(** Terms are stable by substitution *)

Lemma subst_term : forall t z u,
  term u -> term t -> term ([z ~> u]t).
Proof.
  induction 2; simpls*.
  case_var*.
  apply_fresh term_abs as y. rewrite* subst_open_var.
Qed.

Hint Resolve subst_term.

(** Values are stable by substitution *)

Lemma subst_value : forall t z u,
  term u -> value t -> value ([z ~> u]t).
Proof.
  introv HU HV. inverts HV; try solve [ constructors* ].
  { lets: subst_term z HU H. applys* value_abs. }
Qed.

(* ********************************************************************** *)
(** ** Terms are stable through open *)

(** Conversion from locally closed abstractions and bodies *)

Lemma term_abs_to_body : forall t1,
  term (trm_abs t1) -> body t1.
Proof.
  intros. unfold body. inversion* H.
Qed.

Lemma body_to_term_abs : forall t1,
  body t1 -> term (trm_abs t1).
Proof.
  intros. inversion* H.
Qed.

Hint Resolve term_abs_to_body body_to_term_abs.

(** ** Opening a body with a term gives a term *)

Lemma open_term : forall t u,
  body t -> term u -> term (t ^^ u).
Proof.
  intros. destruct H. pick_fresh y. rewrite* (@subst_intro y).
Qed.

Hint Resolve open_term.


(* ********************************************************************** *)
(** ** Well-formedness of store *)

Lemma sto_ok_binds : forall mu l t,
   sto_ok mu -> binds l t mu -> term t.
Proof.
  introv. gen mu. (*todo remove: applys env_ind; [ introv Ok B | introv IH Ok B ].*)
  induction mu as [|mu l1 t1] using env_ind; introv Ok B.
  false* binds_empty_inv.
  inverts Ok. false* empty_push_inv.
   forwards (?&?&?): (eq_push_inv (rm H)). subst.
   binds_cases B; subst~. (* todo: could simplify *)
Qed.

Hint Resolve sto_ok_binds.


(* ********************************************************************** *)
(** ** Regularity of relations *)

(** A typing relation holds only if the environment has no
   duplicated keys and the pre-term is locally-closed. *)

Lemma typing_regular : forall E Y e T,
  typing E Y e T -> ok E /\ term e.
Proof.
  split; induction H; autos*.
  pick_fresh y; forwards~ K: (H0 y).
Qed.

(** The value predicate only holds on locally-closed terms. *)

Lemma value_regular : forall e,
  value e -> term e.
Proof.
  induction 1; autos*.
Qed.

(** A reduction relation only holds only on well-formed objects. *)

Lemma red_regular : forall c c',
  red c c' ->
     (term (fst c) /\ term (fst c'))
  /\ (sto_ok (snd c) /\ sto_ok (snd c')).
Proof.
  lets: value_regular. induction 1; simpl; jauto.
Qed.

(** Automation for reasoning on well-formedness. *)

Hint Extern 1 (ok ?E) =>
  match goal with
  | H: typing E _ _ _ |- _ => apply (proj1 (typing_regular H))
  end.

Hint Extern 1 (term ?t) =>
  match goal with
  | H: typing _ _ t _ |- _ => apply (proj2 (typing_regular H))
  | H: red (t,_) _ |- _ => apply (proj1 (proj1 (red_regular H)))
  | H: red _ (t,_) |- _ => apply (proj2 (proj1 (red_regular H)))
  | H: value t |- _ => apply (value_regular H)
  end.

Hint Extern 1 (sto_ok ?mu) =>
  match goal with
  | H: red (_,mu) _ |- _ => apply (proj1 (proj2 (red_regular H)))
  | H: red _ (_,mu) |- _ => apply (proj2 (proj2 (red_regular H)))
  | H: sto_typing _ mu |- _ => apply (proj1 H)
  end.

Hint Constructors typing.

Implicit Types Y : phi.
Implicit Types P Q : conf->Prop.


(* ********************************************************************** *)
(** ** Property of [val_or_var] *)

(** [val_of_var] is stable by substitution. *)

Lemma subst_val_or_var : forall t z u,
  value u -> val_or_var t -> val_or_var ([z ~> u]t).
Proof using.
  introv HT [HV|(x&->)].
  { left. applys* subst_value. }
  { simpl. case_var. { left*. } { right*. } }
Qed.

Hint Extern 1 (val_or_var ([_ ~> _]_)) => apply subst_val_or_var.


(** Values and variables don't reduce. *)

Lemma red_val_or_var_inv : forall t mu c,
  val_or_var t ->
  (t,mu) --> c ->
  False.
Proof using.
  introv Ht Hc. inverts Ht as.
  { introv Hv. inverts Hv; inverts Hc. }
  { introv (x&->). inverts Hc. }
Qed.

(** Values-or-variables well-typed in the empty contexts are values. *)

Lemma value_of_not_var : forall t S T,
  typing empty S t T ->
  val_or_var t ->
  value t.
Proof using.
  introv HT [Hv|(x&->)].
  { auto. }
  { inverts HT as. intros K E. false binds_empty_inv E. }
Qed.


(* ########################################################### *)
(* ########################################################### *)
(* ########################################################### *)
(** * Definition of type soundness *)

(** Well-formed configurations *)

Definition conf_term (c:conf) : Prop :=
  let '(t,mu) := c in term t /\ sto_ok mu.

(** Transitive closure of the reduction relation *)

Inductive reds : conf -> conf -> Prop :=
  | reds_here : forall c,
      conf_term c ->
      reds c c
  | reds_step : forall c1 c2 c3,
      red c1 c2 ->
      reds c2 c3 ->
      reds c1 c3.

(** Terminal configurations *)

Definition conf_value (c:conf) : Prop :=
  let '(t,mu) := c in value t.

Hint Unfold conf_value.

(** Safety for all executions of a configuration *)

Definition safe (c:conf) : Prop :=
  forall c', reds c c' -> conf_value c' \/ (exists c'', c' --> c'').

(** Well-typed configurations *)

Definition conf_well_typed (c:conf) : Prop :=
  let '(t,mu) := c in
  exists Y T,
     empty ! Y |= t ~: T
  /\ Y |== mu.

(** Type soundness: all well-typed configurations execute safely *)

Definition type_soundness :=
  forall c, conf_well_typed c -> safe c.


(* ########################################################### *)
(* ########################################################### *)
(* ########################################################### *)
(** * Weakening, Substitution, and Extension lemmas *)

(** Typing is preserved by weakening. *)

Lemma typing_weaken : forall G E F Y t T,
   (E & G) ! Y |= t ~: T ->
   ok (E & F & G) ->
   (E & F & G) ! Y |= t ~: T.
Proof.
  introv Typ. gen_eq H: (E & G). gen G.
  induction Typ; introv EQ Ok; subst.
  apply* typing_var. apply* binds_weaken.
  apply_fresh* typing_abs as y. apply_ih_bind* H0.
  autos*.
  autos*.
  autos*.
  autos*.
  autos*.
  autos*.
  autos*.
  autos*.
Qed.

(** Typing is preserved by substitution. *)

Lemma typing_subst : forall F E Y t T z u U,
  value u ->
  (E & z ~ U & F) ! Y |= t ~: T ->
  E ! Y |= u ~: U ->
  (E & F) ! Y |= [z ~> u]t ~: T.
Proof.
  introv Hu Typt Typu. gen_eq G: (E & z ~ U & F). gen F.
  induction Typt; introv Equ; subst; simpl subst.
  case_var.
    binds_get H0. apply_empty* typing_weaken.
    binds_cases H0; apply* typing_var.
  apply_fresh typing_abs as y.
   rewrite* subst_open_var. apply_ih_bind* H0.
  autos*.
  autos*.
  autos*.
  autos*.
  autos*.
  eauto 9.
  eauto 9.
  eauto 9.
Qed.

(** Typing is preserved by an extension of store typing. *)

Lemma typing_stability : forall Y E Y' t T,
  E ! Y |= t ~: T ->
  extends Y Y' ->
  E ! Y' |= t ~: T.
Proof.
  introv Typ Ext. induction* Typ.
Qed.

Hint Resolve typing_stability.

(** Store typing preserved by allocation of a well-typed term. *)

Lemma sto_typing_push : forall Y mu l t T,
  Y |== mu ->
  empty ! Y |= t ~: T ->
  l # mu -> l # Y ->
  (Y & l ~ T) |== (mu & l ~ t).
Proof.
  unfold sto_typing. introv [StoOk [Dom Ext]]. splits 3.
    auto.
    intros l0 Fr. simpl_dom. lets: (Dom l0).
      asserts* Fr2: (l <> l0). asserts* Fr3: (l0 # Y). auto.
    intros l' T' Has. binds_cases Has.
      destruct (Ext _ _ B) as [t' [Hast' Typt']].
       exists* t'.
      subst. exists* t.
Qed.


(* ########################################################### *)
(* ########################################################### *)
(* ########################################################### *)
(** * Standard Preservation and Progress Proof *)

(** A simple short-hand to help clarifying the following proof.
  It simply destruct the induction hypothesis into smaller pieces. *)

Ltac pres H t' mu' :=
  let Y' := fresh "Y'" in
  let Ext := fresh "Ext" in
  let Typt' := fresh "Typt'" in
  let TypSto' := fresh "TypSto'" in
  destruct~ (H (@refl_equal env empty) t' mu')
                  as [Y' [Ext [Typt' TypSto']]].

(** Preservation (typing is preserved by reduction). *)

Lemma preservation_result : forall Y t t' mu mu' T,
  empty ! Y |= t ~: T ->
  (t,mu) --> (t',mu') ->
  Y |== mu ->
  exists Y',
     extends Y Y'
  /\ empty ! Y' |= t' ~: T
  /\ Y' |== mu'.
Proof.
  introv Typ. gen t' mu'. gen_eq E: (empty : env).
  induction Typ; intros EQ t' mu' Red TypSto; subst;
   inversions Red. (* todo env_fix.*)
  exists Y. inversions Typ1. splits~ 3.
    pick_fresh x. rewrite* (@subst_intro x).
     apply_empty* typing_subst.
  pres IHTyp2 t2' mu'. exists* Y'.
  exists (Y & l ~ T).
   asserts Fr: (l # Y). lets: (proj32 TypSto). auto.
   splits~ 3. apply* sto_typing_push.
  exists Y. splits~ 3.
    inversions Typ.
     destruct~ ((proj33 TypSto) l T) as [t [Hast Typt]].
     rewrites~ (>> binds_functional  Hast).
  exists Y. inversions Typ1. splits~ 3.
    destruct TypSto as [StoOk [Dom Map]]. splits~ 3.
     intros. tests: (l = l0).
       exists t2. split~. rewrites~ (>> binds_functional H8).
       destruct (Map _ _ H1) as [t [Has Typ]]. exists* t.
  autos*.
Qed.

(** Progression (a well-typed term is either a value or it can
  take a step of reduction). *)

Lemma progress_result : forall Y t mu T,
  empty ! Y |= t ~: T ->
  Y |== mu ->
     value t
  \/ exists t', exists mu', (t,mu) --> (t',mu').
Proof.
  introv Typ. gen_eq E: (empty : env). lets Typ': Typ.
  induction Typ; intros; subst.
  false* binds_empty_inv.
  left*.
  right. forwards* V1: value_of_not_var t1.
   destruct~ IHTyp2 as [Val2 | [t2' [mu' Red2]]].
     inversions Typ1; inversions V1. exists* (t0 ^^ t2) mu.
     exists* (trm_app t1 t2') mu'.
  left*.
  left*.
  left*.
  right. forwards* V1: value_of_not_var t1.
    destruct (var_fresh (dom mu)) as [l Fr].
     exists* (trm_loc l) (mu & l ~ t1).
  right. forwards* V1: value_of_not_var t1.
    inversions V1; inversions Typ.
     destruct ((proj33 H0) _ _ H6) as [t' [Has' Typt']].
     exists* t' mu.
  right. forwards* V1: value_of_not_var t1.
   forwards* V2: value_of_not_var t2.
   inversions V1; inversions Typ1.
     destruct ((proj33 H1) _ _ H7) as [t' [Has' Typt']].
     exists* trm_unit (mu & l ~ t2).
  right. forwards* V1: value_of_not_var t1.
    inverts V1; inverts Typ. exists* (trm_int 0) mu.
     constructors*. unfold max. case_if; nat_math.
Qed.

(** "Generic" proof that final type soundness is derived
   from preservation and progress. *)

Lemma small_step_type_soundness :
  type_soundness.
Proof using.
  intros (t,mu) (Y&T&HT&HS). introv R.
  gen_eq c: (t,mu). gen Y t mu. induction R; intros; subst.
  { intros. lets [Hv|(t'&mu'&Hp)]: progress_result HT HS.
    { left. hnfs*. }
    { right. exists*. } }
  { intros. destruct c2 as (t2,mu2).
    forwards* (Y'&EP'&HT'&HS'): preservation_result HT H. }
Qed.


(* ########################################################### *)
(* ########################################################### *)
(* ########################################################### *)
(** * Omni-Small-Step Soundness Proof *)

(** Omni-Small-Step semantics [omnismall c Y]. *)

CoInductive omnismall : conf -> (conf->Prop) -> Prop :=
  | omnismall_beta : forall mu t1 t2 P,
      sto_ok mu ->
      term (trm_abs t1) ->
      value t2 ->
      P (t1 ^^ t2, mu) ->
      omnismall (trm_app (trm_abs t1) t2, mu) P
  | omnismall_ref : forall mu t1 P,
      sto_ok mu ->
      value t1 ->
      (forall l, l # mu -> P (trm_loc l, (mu & l ~ t1))) ->
      omnismall (trm_ref t1, mu) P
  | omnismall_get : forall mu l t P,
      sto_ok mu ->
      binds l t mu ->
      P (t, mu) ->
      omnismall (trm_get (trm_loc l), mu) P
  | omnismall_set : forall mu l t2 P,
      sto_ok mu ->
      value t2 ->
      P (trm_unit, mu & l ~ t2) ->
      omnismall (trm_set (trm_loc l) t2, mu) P
  | omnismall_rand : forall mu n P,
      (forall m, 0 <= m < max n 1 -> P ((trm_int m), mu)) ->
      omnismall (trm_rand (trm_int n), mu) P
  | omnismall_app_2 : forall mu t1 t2 P1 P,
      value t1 ->
      omnismall (t2, mu) P1 ->
      (forall t2' mu', P1 (t2',mu') -> P (trm_app t1 t2', mu')) ->
      omnismall (trm_app t1 t2, mu) P.

Hint Constructors omnismall typing.

(** Postcondition associated with a type *)

Definition st_for_typ (Y:phi) (T:typ) : conf->Prop :=
  fun '(t,mu) =>
    exists Y', extends Y Y'
             /\ Y' |== mu
             /\ empty ! Y' |= t ~: T.

Lemma st_for_typ_current : forall Y mu t T,
  Y |== mu ->
  empty ! Y |= t ~: T ->
  st_for_typ Y T (t,mu).
Proof using.
  introv HS HT. exists Y. splits*.
Qed.

Hint Resolve st_for_typ_current.

Lemma st_for_typ_ref : forall l mu t Y T,
  Y |== mu ->
  value t ->
  l # mu ->
  empty ! Y |= t ~: T ->
  st_for_typ Y (typ_ref T) (trm_loc l, mu & l ~ t).
Proof using.
  introv HY HV Fr HT. exists (Y & l ~ T).
  lets Fr': (proj32 HY) Fr.
  splits*. { applys* sto_typing_push. }
Qed.

Lemma st_for_typ_set : forall l mu t Y T,
  Y |== mu ->
  value t ->
  empty ! Y |= t ~: T ->
  binds l T Y ->
  st_for_typ Y typ_unit (trm_unit, mu & l ~ t).
Proof using.
  introv HY HV HT Hl. exists Y. splits*.
  destruct HY as [StoOk [Dom Map]]. splits~ 3.
  intros. tests: (l = l0).
    exists t. split~. rewrites~ (>> binds_functional Hl).
    destruct (Map _ _ H) as [t' [Has Typ]]. exists* t'.
Qed.

(** Soundness proof *)

Definition omnismall_soundness :=
  forall mu t Y T,
  Y |== mu ->
  empty ! Y |= t ~: T ->
  value t \/ omnismall (t,mu) (st_for_typ Y T).

Lemma omnismall_soundness_induction :
  omnismall_soundness.
Proof using.
  introv HS HT. gen_eq E: (@empty typ). gen mu.
  induction HT; intros; subst; try solve [left* ].
  { false* binds_empty_inv. }
  { left*. constructors.
    { applys term_abs L. intros x Hx. forwards*: H x. } }
  { right. forwards* V1: value_of_not_var t1.
    forwards [V2|HR2]: IHHT2. { eauto. } { auto. }
    { inverts V1; inverts HT1. applys* omnismall_beta.
      applys* st_for_typ_current.
      pick_fresh x. rewrite* (@subst_intro x).
     apply_empty* typing_subst. }
    { applys* omnismall_app_2. introv (Y'&EY'&TS1&TT1). exists* Y'. } }
  { right. forwards* V1: value_of_not_var t1.
    applys* omnismall_ref. intros l Hl. applys* st_for_typ_ref. }
  { right. forwards* V1: value_of_not_var t1.
    inverts V1; inverts HT.
    destruct~ ((proj33 HS) l0 T) as [t [Hast Typt]].
    applys* omnismall_get. }
  { right. forwards* V1: value_of_not_var t1.
    forwards* V2: value_of_not_var t2.
    inverts V1; inverts HT1.
    applys* omnismall_set. applys* st_for_typ_set. }
  { right. forwards* V1: value_of_not_var t1.
    inverts V1; inverts HT. applys* omnismall_rand. }
Qed.

(** We next give a generic proof that [omnismall_soundness_induction]
    entails [type soundness], under the assumption that the omni-small-step
    judgment admits the expected interpretation w.r.t. the standard
    small-step judgment. *)

Parameter omnismall_characterization : forall c P,
  omnismall c P <->
    (   (exists c', c --> c')
     /\ (forall c', c --> c' -> P c')).

(** Statement of the soundness property using omni-small-step. *)

Lemma soundness_of_omnismall_soundness :
  omnismall_soundness ->
  type_soundness.
Proof using.
  introv Sou. intros (t,mu) (Y&T&HT&HY). introv HR.
  gen_eq c: (t,mu). gen t mu Y T. induction HR; intros; subst.
  { intros. forwards [HV|M]: Sou HY HT.
    { left*. }
    { right. rewrite* omnismall_characterization in M. } }
  { lets [HV|M]: Sou HY HT.
    { inverts HV; inverts H. }
    { rewrite omnismall_characterization in M. destruct M as (M1&M2).
      specializes M2 H. destruct c2 as (t2,mu2).
      destruct M2 as (Y'&EY'&HY'&HT'). applys* IHHR HT'. } }
Qed.


(* ########################################################### *)
(* ########################################################### *)
(* ########################################################### *)
(** * Omni-Big-Step Soundness Proof *)

(* ########################################################### *)
(** ** Entailment on postconditions *)

Definition himpl (H1 H2:conf->Prop) : Prop :=
  forall h, H1 h -> H2 h.

Notation "H1 ==> H2" := (himpl H1 H2) (at level 55).

Lemma himpl_trans : forall H2 H1 H3,
  (H1 ==> H2) ->
  (H2 ==> H3) ->
  (H1 ==> H3).
Proof using. introv M1 M2. unfolds* himpl. Qed.


(* ########################################################### *)
(** ** Definition of Omni-Big-Step *)

(** Judgment [omnibig c Q]. Here [Q] is a predicate over configurations,
    but in practice in contains only values. *)

CoInductive coomnibig : conf -> (conf->Prop) -> Prop :=
  | coomnibig_val : forall c Q,
      conf_value c ->
      Q c ->
      coomnibig c Q
  | coomnibig_beta : forall mu t1 t2 Q,
      sto_ok mu ->
      term (trm_abs t1) ->
      value t2 ->
      coomnibig (t1 ^^ t2, mu) Q ->
      coomnibig (trm_app (trm_abs t1) t2, mu) Q
  | coomnibig_rand : forall mu n Q,
      (forall m, 0 <= m < max n 1 -> Q ((trm_int m), mu)) ->
      coomnibig (trm_rand (trm_int n), mu) Q
  | coomnibig_ref : forall mu t1 Q,
      sto_ok mu ->
      value t1 ->
      (forall l, l # mu ->
        Q (trm_loc l, (mu & l ~ t1))) ->
      coomnibig (trm_ref t1, mu) Q
  | coomnibig_get : forall mu l t Q,
      sto_ok mu ->
      binds l t mu ->
      Q (t, mu) ->
      coomnibig (trm_get (trm_loc l), mu) Q
  | coomnibig_set : forall mu l t2 Q,
      sto_ok mu ->
      value t2 ->
      Q (trm_unit, mu & l ~ t2) ->
      coomnibig (trm_set (trm_loc l) t2, mu) Q
  | coomnibig_app_2 : forall mu t1 t2 Q1 Q,
      value t1 ->
      ~ value t2 ->
      coomnibig (t2, mu) Q1 ->
      (forall t2' mu', Q1 (t2',mu') -> coomnibig (trm_app t1 t2', mu') Q) ->
      coomnibig (trm_app t1 t2, mu) Q.

(** [coomnibig] is covariant in the postcondition (omni-big-conseqence) *)

Lemma coomnibig_conseq : forall c Q1 Q2,
  coomnibig c Q1 ->
  Q1 ==> Q2 ->
  coomnibig c Q2.
Proof using.
  introv M W. unfolds himpl. gen c Q1 Q2. cofix IH.
  intros. inverts M; try solve [ constructors* ].
  { applys* coomnibig_get. }
  { applys* coomnibig_app_2. }
Qed.

(** We reuse [st_for_typ] from the previous section, and
    add two lemmas. *)

Lemma extends_trans : forall Y2 Y1 Y3,
  extends Y1 Y2 ->
  extends Y2 Y3 ->
  extends Y1 Y3.
Proof using. unfolds extends. autos*. Qed.

Lemma st_for_typ_extends : forall Y Y' T,
  extends Y Y' ->
  (st_for_typ Y' T) ==> (st_for_typ Y T).
Proof using.
  introv EY. intros (t,mu) (Y''&EY'&Hs&Hv).
  exists Y''. splits*. { applys* extends_trans Y'. }
Qed.

(** Type soundness proof *)

(** Note: if we try to prove directly the statement:

[[
    Lemma coomnibig_soundness_coinduction : forall t mu Y T,
      Y |== mu ->
      empty ! Y |= t ~: T ->
      coomnibig (t,mu) (st_for_typ Y T).
]]

  then Cpq's guard checker is not happy because we apply the consequence
  rule for [coomnibig] needs to be applied before the coinduction hypothesis.
  Hence, we need to tweak the statement slightly for the coinduction,
  then derive the desired result. *)

Definition coomnibig_soundness := forall t mu Y T Q,
  Y |== mu ->
  empty ! Y |= t ~: T ->
  (st_for_typ Y T ==> Q) ->
  coomnibig (t,mu) Q.

Lemma coomnibig_soundness_coinduction :
  coomnibig_soundness.
Proof using.
  unfolds. cofix IH. introv HS HT HQ. inverts HT as.
  { introv HK HT1. false* binds_empty_inv. }
  { intros L M. applys* coomnibig_val. hnf.
    constructors. applys term_abs L. intros x Hx. forwards*: M x. }
  { introv HV1 M1 M2. forwards V1: value_of_not_var M1 HV1.
    inverts V1; inverts M1. tests C: (value t2).
    { applys* coomnibig_beta. applys* IH. pick_fresh x.
      rewrite* (@subst_intro x). apply_empty* typing_subst. }
    { applys* coomnibig_app_2 (st_for_typ Y T1).
      { applys* IH. hnfs*. }
      { introv (Y'&EY'&HY'&HT'). applys* IH. applys himpl_trans HQ.
        applys st_for_typ_extends EY'. } } }
  { intros K. applys* coomnibig_val. }
  { intros K. applys* coomnibig_val. }
  { intros K B. applys* coomnibig_val. }
  { intros HV1 M1. forwards V1: value_of_not_var M1 HV1.
    applys* coomnibig_ref.
    { intros l Hl. applys HQ. applys* st_for_typ_ref. } }
  { introv HV1 M1. forwards V1: value_of_not_var M1 HV1.
    inverts V1; inverts M1.
    destruct~ ((proj33 HS) l0 T) as [t [Hast Typt]].
    applys* coomnibig_get. }
  { introv HV1 HV2 M1 M2. forwards V2: value_of_not_var M2 HV2.
    forwards V1: value_of_not_var M1 HV1. inverts V1; inverts M1.
    applys* coomnibig_set. applys HQ. applys* st_for_typ_set. }
  { introv HV1 M1. forwards V1: value_of_not_var M1 HV1.
     inverts V1; inverts M1. applys* coomnibig_rand.
     (* details: intros. applys HQ. unfolds. exists Y. splits*. *)  }
Qed.

(** The original statement is a trivial consequence. *)

Lemma coomnibig_soundness_result : forall t mu Y T,
  Y |== mu ->
  empty ! Y |= t ~: T ->
  coomnibig (t,mu) (st_for_typ Y T).
Proof using.
  introv HS HT. applys* coomnibig_soundness_coinduction. hnfs*.
Qed.

(** Note that the statement over which we carried the coinduction
    is equivalent, not stronger. *)

Lemma coomnibig_soundness_coinduction' :
  coomnibig_soundness.
Proof using.
  unfolds. introv HS HT HQ. applys coomnibig_conseq.
  { applys* coomnibig_soundness_result. }
  { auto. }
Qed.

(** We next give a generic proof that [omnibig_soundness_coinduction]
    entails [type soundness]. *)

(** In the file Omnisemantics.v, we proved in lemma [safe_of_coomnibig] that
    [coomnibig c Q] for any [Q] entails [safe Q], for a coinductive-big-step
    definition of safety, which we proved equivalent to the small-step
    characterization of safety used in the present file (lemma [ssafe_eq_safe]).
    To avoid reproducing all the development here, let us simply assume in
    this file the property [safe_of_coomnibig]. *)

Parameter safe_of_coomnibig : forall c Q,
  coomnibig c Q ->
  safe c.

(** We show that the result established by coinduction indeed
    entails type soundess. *)

Lemma soundness_of_coomnibig_soundness :
  coomnibig_soundness ->
  type_soundness.
Proof using.
  intros HS (t,mu) (Y&T&HY&HT). applys* safe_of_coomnibig.
  applys* coomnibig_soundness_result.
Qed.


